#  ********************************************************************************
#
# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reserving QPU resources in instruction scheduling."""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass, field
import enum
from functools import reduce

from iqm.pulse.playlist.instructions import Block, ReadoutTrigger
from iqm.pulse.playlist.schedule import Schedule, Segment


class SchedulingAlgorithm(enum.Enum):
    """Algorithms for resolving composite TimeBoxes into atomic ones."""

    HARD_BOUNDARY = "HARD_BOUNDARY"
    """Respects the ``TimeBox`` boundary such that the longest channel with a box defines
    its boundary and all other channels are padded to this length (using the specified ``SchedulingStrategy``)."""
    TETRIS = "TETRIS"
    """Will pack the schedule as tightly as possible while respecting the defined scheduling neighborhood."""


class SchedulingStrategy(enum.Enum):
    """Different scheduling strategies for the contents of composite TimeBoxes."""

    ASAP = "ASAP"
    """TimeBox contents are scheduled as soon as possible within the box."""
    ALAP = "ALAP"
    """TimeBox contents are scheduled as late as possible within the box."""


@dataclass
class TimeBox:
    """Container for one or more instruction schedule fragments, to be scheduled according to a given strategy.

    Each TimeBox can be labeled using a human-readable *label* describing it, and operates on a number
    of *locus components*, using some of their control channels.  It can be either *atomic* or
    *composite*.

    * An atomic box only contains a single :class:`.Schedule`.

    * A composite box contains a sequence of other TimeBoxes as its children.
      The locus components are the union of the locus components of the children.
      If two children use the same channel so that they cannot happen simultaneously, they must
      happen in the order they occur in the sequence.

    A box can be made atomic by *resolving* it using :class:`.ScheduleBuilder.resolve_timebox`.
    The time duration of the box is determined by its contents and the way they are scheduled during the resolution.

    TimeBoxes can be concatenated with the following rules:

    * Addition concatenates the children of the operands into a single TimeBox.

    * The pipe operation groups two TimeBoxes together without concatenating.

    * Iterables of Boxes are treated as the sum of the elements.

    Let ``a, b, c, d`` be TimeBoxes. Then

    .. code-block:: python

        a_then_b = a + b
        c_then_d = (c + d).set_alap()
        abcd = a_then_b | c_then_d

        abb = a + [b, b]
        ccd = [c, c] | d

        all_together = abcd | abb + ccd

        # is equivalent to:
        all_together = a + b | (c + d).set_alap() | a + b + b + c + (c | d)

    """

    label: str
    """Description the contents of the box for users' convenience. Has no functional effect."""

    locus_components: set[str]
    """Names of the QPU components on which this timebox operates. These can include additional components
    to the ones included in one of the channels occupied by this ``TimeBox``. The components included in this
    attribute will be blocked in scheduling, in addition to the ones dictated by the neighborhood range (see
    :attr:`.neighborhood_components`)."""

    atom: Schedule | None
    """Resolved contents of the TimeBox, or None if not resolved."""

    children: tuple[TimeBox, ...] = field(default_factory=tuple)
    """Further Timeboxes inside this TimeBox."""

    scheduling: SchedulingStrategy = SchedulingStrategy.ASAP
    """Determines how the contents of a composite TimeBox are scheduled by ScheduleBuilder.
    Has no meaning for an atomic TimeBox."""

    scheduling_algorithm: SchedulingAlgorithm = SchedulingAlgorithm.HARD_BOUNDARY
    """Determines the algorithm used in converting the TimeBox to a Schedule."""

    neighborhood_components: dict[int, set[str]] = field(default_factory=dict)
    """Dict of neighborhood range integers mapped to sets of components neighboring the locus of this ``TimeBox``.
     These are used in the scheduling when the corresponding neighborhood range is used.
     The scheduling algorithm computes the neighborhood components (unless it has been already precomputed by
     e.g. the `GateImplementation`) and caches them under this attribute. Neighborhood range 0 means just the components
     affected by one of the channels in ``self.atom`` + ``self.locus``, 1 means also neighboring couplers, 2 the
     components connected to those couplers, and so on. Note: range 0 may differ from ``self.locus_components``: it can
     have additional components that have occupied channels in ``self`` but are not defined as a part of the 'locus' of
     this ``TimeBox`` for any reason.
    """

    @classmethod
    def composite(
        cls,
        boxes: Iterable[TimeBox | Iterable[TimeBox]],
        *,
        label: str = "",
        scheduling: SchedulingStrategy = SchedulingStrategy.ASAP,
        scheduling_algorithm: SchedulingAlgorithm = SchedulingAlgorithm.HARD_BOUNDARY,
    ) -> TimeBox:
        """Build a composite timebox from a sequence of timeboxes.

        Args:
            boxes: contents of the new timebox. Any iterables of timeboxes will be flattened (recursively) and extended
                to the contents in the same order.
            label: label of the new timebox
            scheduling: scheduling strategy to use when resolving the new timebox
            scheduling_algorithm: scheduling algorithm to use when resolving the new timebox

        Returns:
            composite timebox containing ``boxes`` as its children

        """
        children = []
        for child in boxes:
            if isinstance(child, TimeBox):
                child_box = child
                children.append(child_box)
            else:
                child_box = TimeBox.composite(child, scheduling=scheduling, scheduling_algorithm=scheduling_algorithm)
                children.extend(list(child_box.children))
        if boxes:
            locus_components = set.union(*(box.locus_components for box in children))
        else:
            locus_components = set()
        return cls(
            label=label,
            locus_components=locus_components,
            atom=None,
            children=tuple(children),
            scheduling=scheduling,
            scheduling_algorithm=scheduling_algorithm,
        )

    @classmethod
    def atomic(cls, schedule: Schedule, *, locus_components: Iterable[str], label: str) -> TimeBox:
        """Build an atomic timebox from a schedule.

        Args:
            schedule: contents of the new timebox
            locus_components: names QPU components ``schedule`` operates on
            label: label of the new timebox

        Returns:
            atomic timebox containing ``schedule``

        """
        return cls(label=label, locus_components=set(locus_components), atom=schedule, children=())

    def validate(self, path: tuple[str, ...] = ()) -> None:
        """Validate the contents of the TimeBox.

        Args:
            path: Labels of ancestor boxes, to generate a better error message.

        """
        new_path = path + (self.label,)
        if self.atom:
            self.atom.validate(new_path)
            return

        for child in self.children:
            child.validate(new_path)

    def set_asap(self) -> TimeBox:
        """Set the scheduling strategy to As soon as possible (ASAP)."""
        self.scheduling = SchedulingStrategy.ASAP
        return self

    def set_alap(self) -> TimeBox:
        """Set the scheduling strategy to As late as possible (ALAP)."""
        self.scheduling = SchedulingStrategy.ALAP
        return self

    def __getitem__(self, item: int) -> TimeBox:
        """Shortcut for ``self.children[item]``."""
        if not self.children:
            raise ValueError(f"Tried to access a child of {self}, which is atomic.")
        return self.children[item]

    def _add_children(self, other: TimeBox) -> TimeBox:
        """Concat the children of self and other together."""
        left = self.children if self.atom is None else (self,)
        right = other.children if other.atom is None else (other,)
        return TimeBox(
            label=self.label,
            locus_components=self.locus_components.union(other.locus_components),
            atom=None,
            children=left + right,
            scheduling=self.scheduling,
            scheduling_algorithm=self.scheduling_algorithm,
        )

    def __add__(self, other: TimeBox | Iterable[TimeBox]) -> TimeBox:
        """Return a new TimeBox which has the contents of this and another TimeBox concatenated.

        Used to concatenate multiple TimeBoxes, like atomic operations, to a single logical entity.

        The add operation is associative: for boxes ``a, b, c``, these are equivalent:
        ``a+b+c == (a+b)+c == a+(b+c) == a+[b,c] == [a,b]+c``.

        The scheduling strategy and label are given by `self`, i.e. the leftmost operand.

        Args:
             other: TimeBox or an iterable of TimeBoxes whose contents to merge.

        Returns:
            A new instance containing the children of both boxes.

        """
        if issubclass(type(other), TimeBox) and type(other) is not TimeBox:  # strict subclass
            # allow subclasses to override __add__ such that __radd__ also works consistent with that logic
            return other.__radd__(self)  # type: ignore[union-attr]
        if isinstance(other, TimeBox):
            return self._add_children(other)
        try:
            return reduce(lambda x, y: x + y, other, self)
        except TypeError as err:
            raise TypeError(f"Cannot add a TimeBox and a {type(other)}.") from err

    def __radd__(self, other: TimeBox | Iterable[TimeBox]) -> TimeBox:
        if isinstance(other, TimeBox):
            return other._add_children(self)
        it = iter(other)
        try:
            first = next(it)
        except StopIteration:
            return self
        return reduce(lambda x, y: x + y, it, first) + self

    def __iadd__(self, other: TimeBox | Iterable[TimeBox]) -> TimeBox:
        """Concatenate contents of another TimeBox to this TimeBox.

        Args:
             other: TimeBox whose contents to merge.

        Returns:
            Self, modified to contain the children of both boxes.

        """
        if self.atom is not None:
            raise ValueError("Cannot add content to an atomic TimeBox.")
        new = self + other
        self.children = new.children
        self.locus_components = new.locus_components
        return self

    def __or__(self, other: TimeBox | Iterable[TimeBox]) -> TimeBox:
        """Construct a new composite TimeBox that contains `self` and `other`.

        Used to group two TimeBoxes without mixing their properties.
        Useful for separating boxes which serve a logically distinct purpose.
        For example, for boxes ``a, b, c, d``,  ``a+b|c+d`` results in the content
        ``[[a, b], [c, d]]``, preserving the properties of ``a+b`` and ``c+d``.
        This way, ``a+b`` can be scheduled according to a different strategy than ``c+d``, for example.
        This is in contrast to ``(a+b)+(c+d) == a+b+c+d``, which results in the content ``[a, b, c, d]``.

        This operation is not associative: ``a|b|c != a|(b|c)`` as these result in box contents ``[[a, b], c]``,
        ``[a, [b, c]]``, respectively.

        Args:
             other: TimeBox to append.

        Returns:
            A new TimeBox containing self and `other` as children.

        """
        if isinstance(other, TimeBox):
            other = [other]
        return TimeBox.composite([self, reduce(lambda x, y: x + y, other)])

    def __ror__(self, other: Iterable[TimeBox]) -> TimeBox:
        return TimeBox.composite([reduce(lambda x, y: x + y, other), self])

    def print(self, _idxs: tuple[int, ...] = ()) -> None:
        """Print a simple representation of the contents of this box."""
        location = "".join(f"[{idx}]" for idx in _idxs)
        location = f"{location}:".ljust(12)
        label = self.label or f"(unnamed on {self.locus_components})"
        atomic = " (atomic)" if self.atom else ""
        print(f"{location}{label}{atomic}")
        for i, child in enumerate(self.children):
            child.print(_idxs + (i,))


class MultiplexedProbeTimeBox(TimeBox):
    """A ``TimeBox`` that contains any number of multiplexed readout pulses for probe channels.

    A ``MultiplexedProbeTimeBox``'s atom contains exactly one ``ReadoutTrigger`` for each probe channel.
    """

    def _multiplex(self, other_atom: Schedule) -> dict[str, Segment]:
        new_segments = dict(self.atom.copy().items())  # type: ignore[union-attr]
        for channel, segment in other_atom.items():
            if channel not in new_segments:
                new_segments[channel] = segment
            elif isinstance(segment[0], ReadoutTrigger) and isinstance(new_segments[channel][0], ReadoutTrigger):
                # multiplex the readout triggers together
                new_segments[channel]._instructions[0] = new_segments[channel][0] + segment[0]
            else:
                new_segments[channel].extend(iter(segment))
        return new_segments

    def __add__(self, other: TimeBox | Iterable[TimeBox]) -> TimeBox:
        """Override ``__add__`` for two atomic ``MultiplexedProbeTimeBox`` instances such that ``ReadoutTrigger``s
        belonging to the same probe channel are multiplexed together. Otherwise, behaves exactly like
        ``TimeBox.__add__``, returning a normal ``TimeBox``.
        """
        if isinstance(other, MultiplexedProbeTimeBox) and self.atom is not None and other.atom is not None:
            new_segments = self._multiplex(other.atom)
            locus_components = self.locus_components.union(other.locus_components)
            max_nb = max(
                max(self.neighborhood_components, default=-1),
                max(other.neighborhood_components, default=-1),
            )
            # Combine a neighborhood for the two boxes only if precomputed for both.
            # If a neighborhood is not precomputed for either one, we must leave it empty in the multiplexed result
            # for the scheduler to compute correctly.
            neighborhood_components: dict[int, set[str]] = {}
            for nb in range(max_nb + 1):
                if nb in self.neighborhood_components and nb in other.neighborhood_components:
                    neighborhood_components[nb] = self.neighborhood_components[nb].union(
                        other.neighborhood_components[nb]
                    )
            return type(self)(
                label=f"MultiplexedProbeTimeBox on {locus_components}",
                locus_components=locus_components,
                atom=Schedule(new_segments),
                children=(),
                scheduling=self.scheduling,
                scheduling_algorithm=self.scheduling_algorithm,
                neighborhood_components=neighborhood_components,
            )
        if isinstance(other, ProbeTimeBoxes):
            return self + other[0]
        return super().__add__(other)

    @staticmethod
    def from_readout_trigger(
        readout_trigger: ReadoutTrigger,
        probe_channel: str,
        locus_components: Iterable[str],
        *,
        label: str = "",
        block_channels: Iterable[str] = (),
        block_duration: int = 0,
    ) -> MultiplexedProbeTimeBox:
        """Build an atomic ``MultiplexedProbeTimeBox` from a single ``ReadoutTrigger`` instruction.

        Args:
            readout_trigger: Readout trigger instruction.
            probe_channel: Name of the probe channel to play ``readout_trigger`` in.
            locus_components: Locus components.
            label: Label of the new timebox.
            block_channels: Names of channels to block.
            block_duration: Duration of the required blocking (in samples).

        Returns:
            atomic timebox containing ``readout_trigger`` in the channel ``probe_channel``.

        """
        schedule = {probe_channel: [readout_trigger]}
        for channel in block_channels:
            schedule[channel] = [Block(block_duration)]  # type: ignore[list-item]
        box = MultiplexedProbeTimeBox(
            label=label,
            locus_components=set(locus_components),
            atom=Schedule(schedule, duration=readout_trigger.duration),
            children=(),
            scheduling=SchedulingStrategy.ASAP,
            scheduling_algorithm=SchedulingAlgorithm.HARD_BOUNDARY,
        )
        return box


class ProbeTimeBoxes(list):
    """Allows multiplexing lists of probe TimeBoxes with each other and ``MultiplexedProbeTimeBox``es.

    The first element is a ``MultiplexedProbeTimeBox`` and the second an atomic ``TimeBox`` containing ``Block``
    instructions on probe channels (corresponding to e.g. the integration time).

    If ``A`` is a ``ProbeTimeBoxes`` and ``B`` is a ``ProbeTimeBoxes``, then ``A+B`` is also a ProbeTimeBoxes instance
    with the timings adjusted correctly and the ``MultiplexedProbeTimeBox``es in each multiplexed together.
    If ``A`` is a ``ProbeTimeBoxes`` and ``B`` is a ``MultiplexedProbeTimeBox`` where the probe instructions are
    multiplexed together and the duration is ``max(A[0].duration, B.duration)`` (the addition is symmetric).
    Otherwise (if ``B`` is a list of ``Timebox``es or a non-probe ``TimeBox``), ``A+B`` works like
    ``TimeBox.__add__``.
    """

    # FIXME: does not yet work with ShelvedProbeTimeBox
    # All the measure gates should probably eventually return lists of TimeBoxes (that's the only way to "tetris"
    # schedule stuff) and we should extend that to cover all the use cases, including Shelved.

    def __init__(self, timeboxes: list[TimeBox]):
        super().__init__(timeboxes)
        if len(self) != 2:
            raise ValueError(
                "ProbeTimeBoxes instances must have exactly two elements, the first being the "
                "MultiplexedProbeTimebox of the physical probe channels and the second the "
                " extra Blocks in the probe channels."
            )
        if not isinstance(self[0], TimeBox) or self[0].atom is None or not isinstance(self[0], MultiplexedProbeTimeBox):
            raise ValueError("The first child must be an atomic MultiplexedProbeTimeBox.")
        if not isinstance(self[1], TimeBox) or self[1].atom is None:
            raise ValueError(
                "The second child must be an atomic TimeBox containing the extra Waits in the probe channels."
            )

    def __add__(self, other: TimeBox | list[TimeBox]) -> TimeBox | list[TimeBox]:  # type: ignore[override]
        """Multiplex ``self`` with other ``ProbeTimeBoxes`` instances or with ``MultiplexedProbeTimeBox`` instances."""
        if isinstance(other, ProbeTimeBoxes):
            duration_self = sum(child.atom.duration for child in self)
            duration_probe_self = self[0].atom.duration
            duration_other = sum(child.atom.duration for child in other)
            duration_probe_other = other[0].atom.duration
            extra_wait_duration = max(duration_self, duration_other) - max(duration_probe_self, duration_probe_other)
            extra_wait_channels = [*self[1].atom.channels(), *other[1].atom.channels()]
            probe_box = self[0] + other[0]
            extra_wait_box = TimeBox.atomic(
                Schedule({ch: Segment([Block(extra_wait_duration)]) for ch in extra_wait_channels}),
                locus_components=set(self[1].locus_components).union(other[1].locus_components),
                label="Virtual probe channel Block",
            )
            if 0 in self[1].neighborhood_components and 0 in other[1].neighborhood_components:
                extra_wait_box.neighborhood_components[0] = (
                    self[1].neighborhood_components[0].union(other[1].neighborhood_components[0])
                )
            return ProbeTimeBoxes([probe_box, extra_wait_box])
        if isinstance(other, MultiplexedProbeTimeBox):
            return self[0] + other
        return list(self) + other



#  ********************************************************************************
#    _____  ____ _
#   / _ \ \/ / _` |  Framework for control
#  |  __/>  < (_| |  and measurement of
#   \___/_/\_\__,_|  superconducting qubits
#
#  Copyright (c) 2019-2025 IQM Finland Oy.
#  All rights reserved. Confidential and proprietary.
#
#  Distribution or reproduction of any information contained herein
#  is prohibited without IQM Finland Oy’s prior written permission.
#  ********************************************************************************
"""Waveform definitions for Fourier Ansatz Spectrum Tuning (FAST) DRAG pulse based on :cite:`Hyyppa_2024`."""

from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache

import numpy as np
from scipy.integrate import simpson

from iqm.pulse.playlist.waveforms import Waveform


def fourier_transform_of_cos_basis_functions_as_tensor(
    n_arr: np.ndarray, frequencies: np.ndarray, pulse_duration: float
) -> np.ndarray:
    r"""Evaluate Fourier transforms of cosine basis functions for given basis function indices and frequencies.

    The nth cosine basis function is given by :math:`g_n(t) = 1 -  \cos(2 \pi n t/t_p)` defined on the interval
    :math:`[0, t_p]`, where :math:`t_p` is the pulse duration. The Fourier transform can be analytically evaluated, see
    Eq. (A7) of :cite:`Hyyppa_2024`. We evaluate the Fourier transform for the basis function indices given by ``n_arr``
    and for the frequencies given by ``frequency_arr``. We store the Fourier transforms as a tensor of dimension
    ``1+dim(frequency_arr)``, such that the first dimension corresponds to the basis function indices, and the following
    dimensions to those of ``frequency_arr``. Thus, the Fourier transform is evaluated essentially for the cartesian
    product of ``n_arr`` and ``frequency_arr``.


    Args:
        n_arr: 1d array of basis function indices, running typically from 1 to N, where N is the
            number of considered basis functions
        frequencies: N-dimensional array of frequencies (in Hz), at which the Fourier transform is evaluated.
            For the computation of FAST DRAG coefficients, ``N=2``.
        pulse_duration: Pulse duration (in s), without zero padding.

    Returns:
        Array containing the Fourier transform as a tensor with a shape ``(len(n_arr), *frequency_arr.shape)``

    """
    # Note np.sinc = sin(pi*x)/(pi*x)
    # Shape product of frequencies and pulse duration to have same number of dimensions as output tensor.
    # Frequencies will be copied across different basis function indices
    f_arr_tp = pulse_duration * frequencies[None]
    # Shape n_arr to have same number of dimensions as output tensor. Basis function indices will be copied
    # across frequencies
    n_arr_shaped = n_arr[:, *[None] * len(frequencies.shape)]
    fourier_transform_tensor = pulse_duration * (
        np.exp(-1j * np.pi * f_arr_tp) * np.sinc(f_arr_tp)
        - 0.5 * np.exp(1j * np.pi * (n_arr_shaped - f_arr_tp)) * np.sinc(n_arr_shaped - f_arr_tp)
        - 0.5 * np.exp(-1j * np.pi * (n_arr_shaped + f_arr_tp)) * np.sinc(n_arr_shaped + f_arr_tp)
    )
    return fourier_transform_tensor


def compute_matrix_of_summed_fourier_transform_inner_products(
    n_arr: np.ndarray,
    weights: np.ndarray,
    suppressed_freq_ranges_2darr: np.ndarray,
    pulse_duration: float,
    time_scaling_factor: float,
    n_points_for_integration: int = 60,
) -> np.ndarray:
    r"""Evaluate matrix :math:`A` defined in Eq. (5) of :cite:`Hyyppa_2024`.

    The matrix element :math:`A_{nm}` is defined as

    .. math:: A_{nm} = \sum_{j=1}^k w_j \int_{f_{l,j}}^{f_{h,j}} \hat{g}_n(f) \hat{g}_m(f)^* \mathrm{d} f,

    where n,m are the row and column indices starting from 1 and extending to :math:`N` denoting the
    number of basis functions, :math:`f_{l,j}` and :math:`f_{h,j}` denote the starting and ending frequency of the
    j'th frequency range to suppress, :math:`\hat{g}_n(f)` denotes the Fourier transform of
    :math:`g_n(f) = 1 -  \cos(2 \pi n t/t_p)` with :math:`t_p` being the pulse duration.

    Importantly, the computations are fully vectorized for fast evaluation.

    Args:
        n_arr: 1d array containing indices of basis functions used in the series expansion,
            typically ranging from 1 to N
        weights: 1d array of weights corresponding to the suppressed frequency ranges
        suppressed_freq_ranges_2darr: 2d array describing the frequency ranges (in Hz), over which we want to suppress
            the Fourier transform. The array should have two columns, the first giving the starting frequency
            for each range and the second giving the ending frequency for each range.
            Note that frequencies will be symmetrically suppressed around the center frequency.
        pulse_duration: Pulse duration (in s) without zero padding.
        time_scaling_factor: Scaling factor for time to avoid excessively large or small values in the matrix. Scaling
            factor of frequency is obtained as the inverse.
        n_points_for_integration: Number of points at which the integrand is evaluated for each interval. At least 50
            points are recommended for the numerical approximation error to be low.

    Returns:
        2d array representing the ``A`` matrix with dimensions ``(len(n_arr), len(n_arr))``

    """
    # Construct 2d array of frequencies at which the Fourier transforms are evaluated when computing the numerical
    # integrals. Here, scaled_freqs_for_all_ranges[j, i] is the ith point of numerical integration in the jth frequency
    # range.
    scaled_freqs_for_all_ranges = np.transpose(
        np.linspace(
            suppressed_freq_ranges_2darr[:, 0] / time_scaling_factor,
            suppressed_freq_ranges_2darr[:, 1] / time_scaling_factor,
            n_points_for_integration,
        )
    )
    # Tensor comprising Fourier transforms \hat{g}_n for the considered basis functions. Here,
    # g_hat_nji_tensor[n, j, i] is the Fourier transform of the n_arr[n]'th basis function
    # evaluated in the i'th integration point of the j'th frequency interval
    g_hat_nji_tensor = fourier_transform_of_cos_basis_functions_as_tensor(
        n_arr, scaled_freqs_for_all_ranges, pulse_duration * time_scaling_factor
    )
    # Construct tensor comprising the products \hat{g}_n*conj(\hat{g}_m). Here, g_hat_product_nmji_tensor[n, m, j, i]
    # denotes the product of basis functions n and m evaluated at the i'th integration point of the j'th frequency
    # interval
    g_hat_product_nmji_tensor = g_hat_nji_tensor[:, None] * np.conjugate(g_hat_nji_tensor[None])
    # Frequency step for the different suppressed ranges. Needed for scaling the numerically evaluated integral
    frequency_steps = scaled_freqs_for_all_ranges[:, 1] - scaled_freqs_for_all_ranges[:, 0]
    # integral_nmj_tensor[n, m, j]: integral of \hat{g}_n*conj(\hat{g}_m) across the j'th frequency interval
    integral_nmj_tensor = simpson(g_hat_product_nmji_tensor, dx=1, axis=-1)
    # a_matrix[n, m]: matrix element corresponding to the basis function pair (n, m)
    a_matrix = np.sum(integral_nmj_tensor * weights * frequency_steps, axis=-1)
    return a_matrix


@lru_cache(maxsize=10000)
def solve_fast_coefficients_for_given_weights_and_ranges(
    number_of_cosines: int,
    pulse_duration: float,
    weights_tuple: tuple[float, ...],
    suppressed_freq_ranges_2d_tuple: tuple[tuple[float, ...], ...],
    n_points_for_integration: int = 60,
) -> np.ndarray:
    r"""Solve for optimal coefficients of the basis functions in a FAST DRAG pulse.

    Computes the optimal coefficients :math:`\{c_n\}_{n=1}^N` for a FAST DRAG pulse with :math:`N` basis functions
    such that the pulse spectrum is suppressed across the given frequency ranges according to the provided weights.
    The computation is based on  Eqs. (A5), (A7), (A12) and (A13) of :cite:`Hyyppa_2024`.

    Note that this function essentially computes the mapping from frequency-domain parameters of a FAST DRAG pulse
    to time-domain parameters.

    The results must be cached to allow efficient generation of playlists containing a large number of pulses with
    same parameters, as in RB.

    Args:
        number_of_cosines: Number of cosine basis functions used in the FAST DRAG pulse
        pulse_duration: Pulse duration (in s) without zero padding.
        weights_tuple: 1d tuple of weights for each suppressed frequency interval used in the objective function
        suppressed_freq_ranges_2d_tuple: 2d tuple describing the frequency ranges (in Hz), over which we want to
            suppress the Fourier transform. Each element of the outer tuple should be a tuple of two floats, the first
            setting the starting frequency for the given range and the second setting the ending frequency for the
            given range.
        n_points_for_integration: Number of points at which the integrand is evaluated for each interval. At least 50
            points are recommended for the numerical approximation error to be low.

    Returns:
        Coefficients of the basis functions as a 1d array.

    """
    # Convert tuple args to np arrays
    weights_arr = np.asarray(weights_tuple)
    suppressed_freq_ranges_2d_arr = np.array([list(inner_tuple) for inner_tuple in suppressed_freq_ranges_2d_tuple])
    # Target rotation angle used in the constraint of the FAST optimization problem. This affects the integral of the
    # pulse, but in typical experimental scenario, the pulse is further scaled so we can set anything here.
    rotation_angle = np.pi
    # Times and frequencies are scaled to values on the order of unity for better numerical stability.
    time_scaling_factor = 1 / pulse_duration
    # Construct matrix A_tilde = [A + A^T, -b; b^T, 0] and vector b = [0; rotation_angle/tg] for solving the
    # optimal coefficients
    n_arr = np.linspace(1, number_of_cosines, number_of_cosines)
    a_tilde = np.zeros((number_of_cosines + 1, number_of_cosines + 1), dtype=complex)
    b_vect = np.zeros((number_of_cosines + 1,), dtype=complex)
    ones_vect = np.ones((number_of_cosines,))

    a_mat = compute_matrix_of_summed_fourier_transform_inner_products(
        n_arr,
        weights_arr,
        suppressed_freq_ranges_2d_arr,
        pulse_duration,
        time_scaling_factor,
        n_points_for_integration=n_points_for_integration,
    )
    # Construct A_tilde of Eq. (A13)
    a_tilde[0:number_of_cosines, 0:number_of_cosines] = a_mat + np.transpose(a_mat)
    a_tilde[0:number_of_cosines, number_of_cosines] = -ones_vect
    a_tilde[number_of_cosines, 0:number_of_cosines] = ones_vect
    # Construct b vector
    b_vect[number_of_cosines] = rotation_angle / (pulse_duration * time_scaling_factor)
    c_opt_full = np.linalg.solve(a_tilde, b_vect)
    # Note that c_opt contains also the Lagrangian multiplier lambda as the last element, which we ignore.
    c_opt = c_opt_full[0:number_of_cosines]

    return np.real(c_opt)


def evaluate_fast_drag_i_envelope(
    t_arr: np.ndarray,
    pulse_duration: float,
    coefficients: np.ndarray,
) -> np.ndarray:
    r"""Evaluate I-envelope of a FAST DRAG pulse for given coefficients.

    The I-envelope is defined as :math:`I(t) = \sum_{n=1}^{N} c_n [1 - (-1)^n \cos(2\pi n t/t_p)]`, where :math:`N` is
    the number of cosine terms in the series, :math:`\{c_n\}` are the coefficients, and the pulse is defined on the
    interval :math:`t \in [-t_p/2, t_p/2]`.

    Args:
        t_arr: Array of time points, at which the function is to be evaluated
        pulse_duration: Pulse duration in the same units as t_arr
        coefficients: Coefficients of a FAST DRAG pulse

    Returns:
        I-envelope of a FAST DRAG pulse evaluated at ``t_arr``

    """
    non_zero_indices = np.logical_and(t_arr > -pulse_duration / 2, t_arr < pulse_duration / 2)
    pulse_samples = np.zeros(t_arr.shape)

    cosine_terms = np.arange(1, len(coefficients) + 1)
    pm = 2 * (cosine_terms % 2) - 1
    pulse_samples[non_zero_indices] = np.sum(
        coefficients[:, None]
        * (1 + pm[:, None] * np.cos(2 * np.pi * cosine_terms[:, None] * t_arr[non_zero_indices] / pulse_duration)),
        axis=0,
    )
    return pulse_samples


def evaluate_fast_drag_q_envelope(
    t_arr: np.ndarray,
    pulse_duration: float,
    coefficients: np.ndarray,
) -> np.ndarray:
    r"""Evaluate Q-envelope of FAST DRAG for given coefficients.

    The Q-envelope is defined as :math:`Q(t) = \sum_{n=1}^{N} c_n n (-1)^n \sin(2\pi n t/t_p)]`, where :math:`N` is
    the number of cosine terms in the series, :math:`\{c_n\}` are the coefficients, and the pulse is defined on the
    interval :math:`t \in [-t_p/2, t_p/2]`.

    Args:
        t_arr: Array of time points, at which the function is to be evaluated
        pulse_duration: Pulse duration in the same units as t_arr
        coefficients: Coefficients of a FAST DRAG pulse

    Returns:
        Q-envelope of a FAST DRAG pulse evaluated at ``t_arr``

    """
    non_zero_indices = np.logical_and(t_arr > -pulse_duration / 2, t_arr < pulse_duration / 2)
    pulse_samples = np.zeros(t_arr.shape)
    sine_terms = np.arange(1, len(coefficients) + 1)
    pm = 1 - 2 * (sine_terms % 2)
    pulse_samples[non_zero_indices] = np.sum(
        coefficients[:, None]
        * sine_terms[:, None]
        * pm[:, None]
        * np.sin(2 * np.pi * sine_terms[:, None] * t_arr[non_zero_indices] / pulse_duration),
        axis=0,
    )

    return pulse_samples


@dataclass(frozen=True)
class SuppressedPulse(Waveform):
    r"""Base class for a control pulse using a series expansion to suppress certain frequencies in its envelope spectrum

    The base class describes control pulses, in which the coefficients of the basis functions
    are chosen to suppress specific frequencies or frequency ranges in the frequency spectrum of the pulse envelope.
    Examples include FAST DRAG and HD DRAG.

    The pulse argument ``compute_coefs_from_frequencies`` allows the user to choose whether the basis function
    coefficients are computed from ``suppressed_frequencies`` during the post-initialization of the pulse
    (thus, overriding any pre-computed values in ``coefficients``), or if the (pre-computed) values in ``coefficients``
    are directly used and ``suppressed_frequencies`` are neglected. The classes deriving from this base class should
    implement the post-initialization logic specific to the given pulse.

    Args:
        full_width: Full width of the pulse corresponding to the pulse duration with non-zero amplitude (in s).
        coefficients: Pre-computed coefficients of the series expansion.
        suppressed_frequencies: Frequencies to be suppressed or center frequencies of intervals to be suppressed
            (in Hz).
        compute_coefs_from_frequencies: Boolean value indicating whether we compute the coefficients from the
            suppressed frequencies or whether we use the pre-computed coefficients. If True, ``suppressed_frequencies``
            are used to compute and override ``coefficients`` in the post-initialization. If False, pre-computed
            ``coefficients`` are used, and ``suppressed_frequencies`` are ignored.

    """

    full_width: float
    coefficients: np.ndarray
    suppressed_frequencies: np.ndarray
    compute_coefs_from_frequencies: bool

    @staticmethod
    def _normalize(samples: np.ndarray) -> np.ndarray:
        """Scale the pulse samples to the interval (-1, 1) to avoid clipping at the instruments.

        Args:
            samples: Array of pulse samples

        Returns:
            re-scaled Array of pulse samples

        """
        return np.clip(samples / np.max(np.abs(samples)), -1, 1)


@dataclass(frozen=True)
class FastDrag(SuppressedPulse):
    r"""Base class for IQ components of the Fourier Ansatz Spectrum Tuning (FAST) DRAG pulse.

    The FAST DRAG pulse shapes the I-envelope in the frequency domain to suppress specified frequency intervals
    according to given weights. Furthermore, the Q-envelope is obtained as a derivative of the I-envelope
    similarly to ordinary DRAG. This class represents a base class for an implementation using cosine functions
    as the I-envelope basis functions. See :cite:`Hyyppa_2024` for more details on FAST DRAG.

    Args:
        number_of_cos_terms: Number of cosine terms in the Fourier series expression of the I-component
        suppressed_interval_widths: Widths of the suppressed frequency intervals (in Hz). The last element corresponds
            to the width of a potential cutoff interval.
        weights: Weights corresponding to the suppressed frequency intervals

    """

    number_of_cos_terms: int
    suppressed_interval_widths: np.ndarray
    weights: np.ndarray
    center_offset: float = 0

    def __post_init__(self) -> None:
        """Post initialization."""
        if self.compute_coefs_from_frequencies:
            suppressed_frequency_ranges_2d_arr = np.zeros((len(self.suppressed_frequencies), 2))
            suppressed_frequency_ranges_2d_arr[:, 0] = self.suppressed_frequencies - self.suppressed_interval_widths / 2
            suppressed_frequency_ranges_2d_arr[:, 1] = self.suppressed_frequencies + self.suppressed_interval_widths / 2
            coefficients = solve_fast_coefficients_for_given_weights_and_ranges(
                self.number_of_cos_terms,
                self.full_width,
                tuple(list(self.weights)),
                tuple([tuple(freq_pair) for freq_pair in suppressed_frequency_ranges_2d_arr]),
            )
            object.__setattr__(self, "coefficients", coefficients)

    @staticmethod
    def non_timelike_attributes() -> dict[str, str]:
        return {
            "compute_coefs_from_frequencies": "",
            "coefficients": "",
            "suppressed_frequencies": "Hz",
            "number_of_cos_terms": "",
            "suppressed_interval_widths": "Hz",
            "weights": "",
        }


@dataclass(frozen=True)
class FastDragI(FastDrag):
    r"""I-component of the Fourier Ansatz Spectrum Tuning (FAST) drag pulse.

    The I-envelope is defined as

    .. math:: I(t) = \sum_{n=1}^{N} c_n [1 - \cos(2\pi n t/t_p + n\pi)],

    where :math:`N` is the number of cosine terms in the series, :math:`\{c_n\}` are the coefficients, and
    the pulse is defined on the interval :math:`t \in [-t_p/2, t_p/2]`.
    """

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        pulse_samples = evaluate_fast_drag_i_envelope(
            sample_coords - self.center_offset,
            self.full_width,
            self.coefficients,
        )
        normalized_samples = FastDrag._normalize(pulse_samples)
        return normalized_samples


@dataclass(frozen=True)
class FastDragQ(FastDrag):
    r"""Q-component of the Fourier Ansatz Spectrum Tuning (FAST) drag pulse.

    The Q-envelope is defined as

    .. math:: Q(t) = \sum_{n=1}^{N} c_n n \sin(2\pi n t/t_p + n\pi)],

    where :math:`N` is the number of cosine terms in the series, :math:`\{c_n\}` are the coefficients, and
    the pulse is defined on the interval :math:`t \in [-t_p/2, t_p/2]`.
    """

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        pulse_samples = evaluate_fast_drag_q_envelope(
            sample_coords - self.center_offset,
            self.full_width,
            self.coefficients,
        )
        normalized_samples = FastDrag._normalize(pulse_samples)
        return normalized_samples



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Enumerations for representing a set of important one- and two-qubit gates."""

from __future__ import annotations

from enum import Enum, IntEnum, unique

import numpy as np


@unique
class XYGate(IntEnum):
    """Enumerate all single-qubit Clifford gates whose rotation axis is in the XY plane.

    Members of this enum can be mapped to the corresponding unitary propagator using
    :data:`XYGATE_UNITARIES`.

    Only used in the tomography experiments.
    """

    IDENTITY = 0
    X_90 = 1
    X_180 = 2
    X_M90 = 3
    Y_90 = 4
    Y_180 = 5
    Y_M90 = 6


XYGATE_UNITARIES = {
    XYGate.IDENTITY: np.eye(2, dtype=complex),
    XYGate.X_90: np.array([[1, -1j], [-1j, 1]], dtype=complex) / np.sqrt(2),
    XYGate.X_180: np.array([[0, -1j], [-1j, 0]], dtype=complex),
    XYGate.X_M90: np.array([[1, 1j], [1j, 1]], dtype=complex) / np.sqrt(2),
    XYGate.Y_90: np.array([[1, -1], [1, 1]], dtype=complex) / np.sqrt(2),
    XYGate.Y_180: np.array([[0, -1], [1, 0]], dtype=complex),
    XYGate.Y_M90: np.array([[1, 1], [-1, 1]], dtype=complex) / np.sqrt(2),
}
"""Mapping of XYGates to the corresponding SU(2) matrices"""


@unique
class TwoQubitGate(Enum):
    """Enumerates a subset of two-qubit gates.

    Members of this enum can be mapped to the corresponding unitary propagator using
    the dictionary returned by :data:`TWO_QUBIT_UNITARIES`.
    """

    CZ = 0
    """Controlled-Z gate."""
    ISWAP = 1
    """iSWAP gate."""
    SQRT_ISWAP = 2
    """Square root of the iSWAP gate."""


TWO_QUBIT_UNITARIES = {
    TwoQubitGate.CZ: np.array(np.diag([1, 1, 1, -1]), dtype=float),
    TwoQubitGate.ISWAP: np.array([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]], dtype=complex),
    TwoQubitGate.SQRT_ISWAP: np.array(
        [[1, 0, 0, 0], [0, 1 / np.sqrt(2), 1j / np.sqrt(2), 0], [0, 1j / np.sqrt(2), 1 / np.sqrt(2), 0], [0, 0, 0, 1]],
        dtype=complex,
    ),
}
"""Mapping of TwoQubitGates to the corresponding U(4) matrices"""



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Single-qubit sqrt(X) gate.

The gate is doing pi/2 X gate, with additional Z rotation to correct phase.

.. math::
   R = e^{-i\pi \^{\sigma}_X/4}

It rotates the qubit state in XZ plane (or around Y axis) for 90 degree.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from iqm.pulse.gate_implementation import CompositeGate
from iqm.pulse.gates.prx import PRX_CustomWaveformsSX, PRX_SinglePulse_GateImplementation
from iqm.pulse.playlist.schedule import Schedule

if TYPE_CHECKING:  # pragma: no cover
    from iqm.pulse.timebox import TimeBox


class SXGate(CompositeGate):
    """SX gate implementation based on PRX gate, by limiting the angle to pi / 2."""

    registered_gates = ("prx",)

    def _call(self) -> TimeBox:  # type: ignore[override]
        """Call PRX gate with angle equals to pi / 2."""
        prx_gate: PRX_SinglePulse_GateImplementation | PRX_CustomWaveformsSX = self.build("prx", self.locus)  # type: ignore[assignment]
        pulse = prx_gate(np.pi / 2, 0.0).atom[prx_gate.channel][0]  # type: ignore[union-attr,index]
        return self.to_timebox(Schedule({prx_gate.channel: [pulse]}))



#  ********************************************************************************
#    _____  ____ _
#   / _ \ \/ / _` |  Framework for control
#  |  __/>  < (_| |  and measurement of
#   \___/_/\_\__,_|  superconducting qubits
#
#  Copyright (c) 2019-2025 IQM Finland Oy.
#  All rights reserved. Confidential and proprietary.
#
#  Distribution or reproduction of any information contained herein
#  is prohibited without IQM Finland Oy’s prior written permission.
#  ********************************************************************************
"""Waveform definitions for a higher-derivative (HD) DRAG pulse based on Appendix B of :cite:`Hyyppa_2024`."""

from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache

import numpy as np

from iqm.pulse.playlist.fast_drag import SuppressedPulse

COSINE_COEFFICIENTS_DICT = {
    0: np.array([1]),
    1: np.array([1.33333333, -0.33333333]),
    2: np.array([1.5, -0.6, 0.1]),
    3: np.array([1.6, -0.8, 0.22857143, -0.02857143]),
    4: np.array([1.66666667, -0.95238095, 0.35714286, -0.07936508, 0.00793651]),
    5: np.array([1.71428571, -1.07142857, 0.47619048, -0.14285714, 0.02597403, -0.0021645]),
    6: np.array(
        [
            1.75000000e00,
            -1.16666667e00,
            5.83333333e-01,
            -2.12121212e-01,
            5.30303030e-02,
            -8.15850816e-03,
            5.82750583e-04,
        ]
    ),
    7: np.array(
        [
            1.77777778e00,
            -1.24444444e00,
            6.78787879e-01,
            -2.82828283e-01,
            8.70240870e-02,
            -1.86480186e-02,
            2.48640249e-03,
            -1.55400155e-04,
        ]
    ),
}
"""
Pre-computed coefficients of the cosine terms in the basis envelope (0th derivative). This dictionary contains a 
mapping from the number of suppressed frequencies to the coefficients of the cosine terms computed using the
function ``solve_cosine_coefs_for_hd_drag``. 
"""


def solve_cosine_coefs_for_hd_drag(
    number_of_suppressed_freqs: int,
) -> np.ndarray:
    r"""Solve cosine coefficients of the basis envelope given the number of suppressed frequencies.

    The cosine coefficients :math:`\{d_k\}_{k=1}^{K+1}` define the basis envelope as
    :math:`g(t) = \sum_{k=1}^{K+1} d_k (1 - \cos(2\pi k t/t_p + k\pi))`, where the pulse is defined on the
    interval :math:`t \in [-t_p/2, t_p/2]`.

    Args:
        number_of_suppressed_freqs: Number of suppressed frequencies

    Returns:
        Coefficient array of length ``number_of_suppressed_freqs + 1``

    """
    # The coefficients x_k can be solved from a matrix equation A*x = b. Let's build A
    if number_of_suppressed_freqs == 0:
        return np.array([1.0])
    n_coefs = number_of_suppressed_freqs + 1
    a_mat = np.zeros((n_coefs, n_coefs))
    for idx_row in range(n_coefs):
        for idx_col in range(n_coefs):
            if idx_row < n_coefs - 1:
                a_mat[idx_row, idx_col] = (-1) ** idx_row * (idx_col + 1) ** (2 * (idx_row + 1))
            else:
                a_mat[idx_row, idx_col] = 1

    b_arr = np.zeros((n_coefs,))
    b_arr[-1] = 1
    cosine_coefs_arr = np.linalg.solve(a_mat, b_arr)
    return cosine_coefs_arr


@lru_cache(maxsize=10000)
def solve_hd_drag_coefficients_from_suppressed_frequencies(
    pulse_duration: float,
    suppressed_freq_arr: tuple[float, ...],
) -> np.ndarray:
    r"""Solve coefficients of the derivative terms in a HD DRAG pulse given pulse duration and frequencies to suppress.

    The coefficients :math:`\{\beta_{2n}}_{n=0}^{K}` of the derivative terms are solved using Eq. (B5) of
    :cite:`Hyyppa_2024` assuming that :math:`\beta_0 = 1`. Here, :math:`K` is the number of suppressed frequencies.

    Note that the duration and frequencies must have matching units, e.g., s and Hz, or ns and GHz.

    Args:
        pulse_duration: Pulse duration (in s).
        suppressed_freq_arr: Frequencies to be suppressed relative to the center drive frequency (in Hz).

    Returns:
        Coefficient array of length ``len(suppressed_freq_arr) + 1``

    """
    # The beta coefficients can be solved from a matrix equation A*beta = b according to Eq. (B5). Let's build A
    suppressed_freqs = np.asarray(suppressed_freq_arr)
    n_coefs = len(suppressed_freqs) + 1
    a_mat = np.zeros((n_coefs, n_coefs))
    a_mat[0, 0] = 1
    a_mat[1:, :] = (-1) ** np.arange(n_coefs)[None] * (pulse_duration * suppressed_freqs[:, None]) ** (
        2 * np.arange(n_coefs)[None]
    )
    b_arr = np.zeros((n_coefs,))
    b_arr[0] = 1
    derivative_coefs_arr = np.linalg.solve(a_mat, b_arr)
    return derivative_coefs_arr


def evaluate_nth_derivative_of_basis_envelope(
    t_arr: np.ndarray, pulse_duration: float, cosine_coefs_arr: np.ndarray, n: int
) -> np.ndarray:
    r"""Evaluate nth derivative of the basis envelope for HD DRAG based on a cosine series.

    The basis envelope is given by :math:`g(t) = \sum_{k=1}^{K+1} d_k (1 - \cos(2\pi k t/t_p + k\pi))`, where the
    pulse is defined on the interval :math:`t \in [-t_p/2, t_p/2]`. The returned derivatives are normalized
    via multiplication by :math:`((t_p/(2\pi))^n`, where :math:`n` is the order of the derivative.


    Args:
        t_arr: Array of time points, at which the function is to be evaluated
        pulse_duration: Pulse duration in the same units as t_arr
        cosine_coefs_arr: Coefficients of the cosine terms in the basis envelope
        n: order of derivative

    Returns:
        Array containing the nth derivative of the basis envelope evaluated at ``t_arr``

    """
    non_zero_indices = np.logical_and(t_arr > -pulse_duration / 2, t_arr < pulse_duration / 2)
    t_arr_nonzero = t_arr[non_zero_indices]
    pulse_samples = np.zeros(t_arr.shape)
    n_cos_coefs = len(cosine_coefs_arr)
    if n == 0:
        non_zero_samples = np.sum(
            cosine_coefs_arr[:, None]
            * (
                1
                - (-1) ** np.arange(1, n_cos_coefs + 1)[:, None]
                * np.cos(2 * np.pi * np.arange(1, n_cos_coefs + 1)[:, None] * t_arr_nonzero[None] / pulse_duration)
            ),
            axis=0,
        )
    elif n % 2 == 0:
        non_zero_samples = np.sum(
            cosine_coefs_arr[:, None]
            * (-1) ** (n // 2 + np.arange(n_cos_coefs)[:, None])
            * np.arange(1, n_cos_coefs + 1)[:, None] ** n
            * np.cos(2 * np.pi * np.arange(1, n_cos_coefs + 1)[:, None] * t_arr_nonzero[None] / pulse_duration),
            axis=0,
        )
    else:
        non_zero_samples = np.sum(
            cosine_coefs_arr[:, None]
            * (-1) ** (n // 2 + np.arange(1, n_cos_coefs + 1)[:, None])
            * np.arange(1, n_cos_coefs + 1)[:, None] ** n
            * np.sin(2 * np.pi * np.arange(1, n_cos_coefs + 1)[:, None] * t_arr_nonzero[None] / pulse_duration),
            axis=0,
        )
    pulse_samples[non_zero_indices] = non_zero_samples
    return pulse_samples


def evaluate_hd_drag_i_envelope(
    t_arr: np.ndarray,
    pulse_duration: float,
    derivative_coefs_arr: np.ndarray,
    cosine_coefs_arr: np.ndarray,
) -> np.ndarray:
    r"""Evaluate I-envelope of HD DRAG given the coefficients of the derivative terms and the cosine terms.

    The I-envelope is defined as :math:`I(t) = \sum_{n=0}^{K} \beta_{2n} g^{(2n)}(t)`, where :math:`K` is the number
    of suppressed frequency ranges, :math:`\{beta_{2n}\}` are the coefficients of the derivative terms, and
    :math:`g(t)` is the basis envelope. The pulse is assumed to start at time ``-pulse_duration/2``,
    and end at time ``pulse_duration/2``.

    Args:
        t_arr: Array of time points, at which the function is to be evaluated
        pulse_duration: Pulse duration in the same units as t_arr
        derivative_coefs_arr: Coefficients of the even derivatives
        cosine_coefs_arr: Coefficients of the cosine terms in the basis envelope

    Returns:
        I-envelope of a HD DRAG pulse evaluated at ``t_arr``

    """
    pulse_samples_i = np.zeros(t_arr.shape)
    for idx, derivative_coef in enumerate(derivative_coefs_arr):
        nth_derivative = evaluate_nth_derivative_of_basis_envelope(t_arr, pulse_duration, cosine_coefs_arr, idx * 2)
        pulse_samples_i += derivative_coef * nth_derivative
    return pulse_samples_i


def evaluate_hd_drag_q_envelope(
    t_arr: np.ndarray,
    pulse_duration: float,
    derivative_coefs_arr: np.ndarray,
    cosine_coefs_arr: np.ndarray,
) -> np.ndarray:
    r"""Evaluate Q-envelope of HD DRAG given the coefficients of the derivative terms and the cosine terms.

    The Q-envelope is defined as :math:`Q(t) = \sum_{n=0}^{K} \beta_{2n} g^{(2n+1)}(t)`, where :math:`K` is the number
    of suppressed frequency ranges, :math:`\{beta_{2n}\}` are the coefficients of the derivative terms, and
    :math:`g(t)` is the basis envelope. The pulse is assumed to start at time ``-pulse_duration/2``,
    and end at time ``pulse_duration/2``.

    Args:
        t_arr: Array of time points, at which the function is to be evaluated
        pulse_duration: Pulse duration in the same units as t_arr
        derivative_coefs_arr: Coefficients of the derivatives
        cosine_coefs_arr: Coefficients of the cosine terms in the basis envelope

    Returns:
        Q-envelope of a HD DRAG pulse evaluated at ``t_arr``

    """
    pulse_samples_i = np.zeros(t_arr.shape)
    for idx, derivative_coef in enumerate(derivative_coefs_arr):
        nth_derivative = evaluate_nth_derivative_of_basis_envelope(t_arr, pulse_duration, cosine_coefs_arr, idx * 2 + 1)
        pulse_samples_i += derivative_coef * nth_derivative
    return pulse_samples_i


@dataclass(frozen=True)
class HdDrag(SuppressedPulse):
    r"""Base class for higher-derivative DRAG based on Eqs. (B1) and (B2) of :cite:`Hyyppa_2024`.

    Base class for IQ components of the higher derivative (HD) drag pulse. Depending on the value of
    ``compute_coefs_from_frequencies``, we compute the coefficients from the suppressed frequencies during the
    post-initialization or use pre-computed coefficients of the derivative terms (neglecting the suppressed
    frequencies). See :class:`SuppressedPulse`.

    """

    center_offset: float = 0

    def __post_init__(self) -> None:
        """Post initialization."""
        if self.compute_coefs_from_frequencies:
            coefficients = solve_hd_drag_coefficients_from_suppressed_frequencies(
                self.full_width,
                tuple(list(self.suppressed_frequencies)),
            )
            object.__setattr__(self, "coefficients", coefficients)

    @staticmethod
    def non_timelike_attributes() -> dict[str, str]:
        return {
            "coefficients": "",
            "suppressed_frequencies": "Hz",
            "compute_coefs_from_frequencies": "",
        }


@dataclass(frozen=True)
class HdDragI(HdDrag):
    r"""I-component of the higher derivative (HD) drag pulse.

    The I-component is defined according to Eq. (B1) of :cite:`Hyyppa_2024` ,

    .. math::
         I(t) = \sum_{n=0}^{K} \beta_{2n} g^{(2n)}(t),

    where :math:`\{\beta_{2n}\}_{n=0}^K` are the coefficients of the derivative terms, :math:`K` is the number of
    suppressed frequencies, and :math:`g(t)` is the basis envelope given by

    .. math::
        g(t) = \sum_{k=1}^K d_k (1 - \cos(2 \pi k t/t_p + k\pi)),

    where :math:`d_k` are pre-computed to ensure continuous derivatives up to order :math:`2K + 1`, :math:`t_p`
    denotes the pulse duration, and the pulse is defined across :math:`t \in (-t_p/2, t_p/2)`.

    The sampled pulse is always normalized to have a maximum value slightly below 1.0.
    """

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        n_suppressed_freqs = len(self.coefficients) - 1
        if (cosine_coefs := COSINE_COEFFICIENTS_DICT.get(n_suppressed_freqs)) is not None:
            samples = evaluate_hd_drag_i_envelope(
                sample_coords - self.center_offset, self.full_width, self.coefficients, cosine_coefs
            )
            normalized_samples = HdDrag._normalize(samples)
            return normalized_samples
        raise ValueError("Too many suppressed frequencies. At most 7 is allowed.")


@dataclass(frozen=True)
class HdDragQ(HdDrag):
    r"""Q-component of the higher derivative (HD) drag pulse.

    The Q-component is defined according to Eq. (B1) of :cite:`Hyyppa_2024` ,

    .. math::
         Q(t) = \sum_{n=0}^{K} \beta_{2n} g^{(2n + 1)}(t),

    where :math:`\{\beta_{2n}\}` are the coefficients, and :math:`g(t)` is the basis envelope given by

    .. math::
        g(t) = \sum_{k=1}^K d_k (1 - \cos(2 \pi k t/t_p + k\pi)),

    where :math:`d_k` are pre-computed to ensure continuous derivatives up to order :math:`2K + 1`, :math:`t_p`
    denotes the pulse duration, and the pulse is defined across :math:`t \in (-t_p/2, t_p/2)`.

    The sampled pulse is always normalized to have a maximum value of slightly below 1.0.
    """

    def _sample(self, sample_coords: np.ndarray) -> np.ndarray:
        n_suppressed_freqs = len(self.coefficients) - 1
        if (cosine_coefs := COSINE_COEFFICIENTS_DICT.get(n_suppressed_freqs)) is not None:
            samples = evaluate_hd_drag_q_envelope(
                sample_coords - self.center_offset, self.full_width, self.coefficients, cosine_coefs
            )
            normalized_samples = HdDrag._normalize(samples)
            return normalized_samples
        raise ValueError("Too many suppressed frequencies. At most 7 is allowed.")



#  ********************************************************************************
#
# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions."""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import replace
from typing import get_args, get_origin

import numpy as np

from exa.common.data.parameter import CollectionType, DataType
from iqm.pulse.playlist import IQPulse
from iqm.pulse.playlist.waveforms import Samples


def map_waveform_param_types(type_hint: type) -> tuple[DataType, CollectionType]:
    """Map a python typehint into EXA Parameter's `(DataType, CollectionType)` tuple.

    Args:
        type: python typehint.

    Returns:
        A `(DataType, CollectionType)` tuple

    Raises:
        ValueError: for a non-supported type.

    """
    value_error = ValueError(f"Nonsupported datatype for a waveform parameter: {type_hint}")
    if hasattr(type_hint, "__iter__") and type_hint is not str:
        if type_hint == np.ndarray:
            data_type = DataType.COMPLEX  # due to np.ndarray not being generic we assume complex numbers
            collection_type = CollectionType.NDARRAY
            return (data_type, collection_type)
        if get_origin(type_hint) is list:
            collection_type = CollectionType.LIST
            type_hint = get_args(type_hint)[0]
        else:
            raise value_error
    else:
        collection_type = CollectionType.SCALAR

    if type_hint is float:
        data_type = DataType.FLOAT
    elif type_hint is int:
        data_type = DataType.INT
    elif type_hint is str:
        data_type = DataType.STRING
    elif type_hint is complex:
        data_type = DataType.COMPLEX
    elif type_hint is bool:
        data_type = DataType.BOOLEAN
    else:
        raise value_error
    return (data_type, collection_type)


def normalize_angle(angle: float) -> float:
    """Normalize the given angle to (-pi, pi].

    Args:
        angle: angle to normalize (in radians)

    Returns:
        ``angle`` normalized to (-pi, pi]

    """
    half_turn = np.pi
    full_turn = 2 * half_turn
    return (angle - half_turn) % -full_turn + half_turn


def phase_transformation(psi_1: float = 0.0, psi_2: float = 0.0) -> tuple[float, float]:
    r"""Implement an arbitrary (RZ, PRX, RZ) gate sequence by modifying the parameters of the
    IQ pulse implementing the PRX.

    By commutation rules we have

    .. math::
       RZ(\psi_2) \: PRX(\theta, \phi) \: RZ(\psi_1) = PRX(\theta, \phi+\psi_2) \: RZ(\psi_1 + \psi_2).

    Hence an arbitrary (RZ, PRX, RZ) gate sequence is equivalent to (RZ, PRX) with adjusted angles.

    Use case: with resonant driving, the PRX gate can be implemented using an :class:`IQPulse` instance,
    and the preceding RZ can be handled by decrementing the local oscillator phase beforehand (something
    the IQPulse instruction can also do), which is equivalent to rotating the local computational frame
    around the z axis in the opposite direction of the required quantum state rotation.

    Args:
        psi_1: RZ angle before the PRX (in rad)
        psi_2: RZ angle after the PRX (in rad)

    Returns:
        change to the PRX phase angle (in rad),
        phase increment for the IQ pulse that implements the remaining RZ (in rad)

    """
    return psi_2, -(psi_1 + psi_2)


def modulate_iq(pulse: IQPulse) -> np.ndarray:
    """Sampled baseband waveform of an IQ pulse.

    Note that :attr:`IQPulse.phase_increment` has no effect on the sampled waveform.
    The upconversion oscillator phase incrementation is a separate action performed by the AWG
    that also affects future IQPulses, and thus cannot be represented by an array of waveform samples.
    To replicate the effect of ``pulse`` on an AWG, one should first perform the increment and then
    play the returned samples.

    Args:
        pulse: IQ pulse.

    Returns:
        The waveform of ``pulse`` as an array of complex-valued samples.

    """
    # TODO: Could be an IQPulse method
    wave = pulse.wave_i.sample() * pulse.scale_i + 1j * pulse.wave_q.sample() * pulse.scale_q
    # starting times of the samples, in units of inverse sample rate
    wave_sampletimes = np.arange(len(wave))
    wave *= np.exp(2j * np.pi * pulse.modulation_frequency * wave_sampletimes + 1j * pulse.phase)
    return wave


def fuse_iq_pulses(iq_pulses: Iterable[IQPulse]) -> IQPulse:
    """Fuse multiple IQPulses into one by concatenating the sampled waveforms.

    Works by flushing :attr:`IQPulse.phase_increment` s to the front, updating the :attr:`IQPulse.phase` s,
    sampling the pulses, concatenating, normalizing the amplitudes, and putting the result
    into a new IQPulse instruction with a ``phase_increment`` that is a sum of the individual ``phase_increment`` s.

    Additionally, to conserve waveform memory on the AWGs, we normalize the waveform phase by setting
    :attr:`IQPulse.phase` of the fused pulse to the flushed phase of the first pulse.

    Args:
        iq_pulses: IQPulse instructions to fuse.

    Returns:
        Fused IQPulse that behaves indentically to the sequence ``iq_pulses`` on an AWG.

    """
    # flush the phase increments to the start of the pulse sequence
    phases = np.array([i.phase for i in iq_pulses])
    phase_increments = np.array([i.phase_increment for i in iq_pulses])

    # flushed_phases[k] == phases[k] - np.sum(phase_increments[k+1:])
    flushed_phases = phases - np.cumsum(phase_increments[::-1])[::-1] + phase_increments

    # Phase normalization of the samples to save waveform memory: There is an internal degree of freedom
    # in the sampled IQPulse: IQPulse.phase can be represented in the global phase of the samples.
    # Fix this d.o.f. by setting the phase of the fused IQ pulse to the phase of the first constituent IQ pulse.
    fused_phase = flushed_phases[0]
    flushed_phases -= fused_phase
    flushed_iq_pulses = [
        replace(instr, phase=phase, phase_increment=0.0) for instr, phase in zip(iq_pulses, flushed_phases)
    ]
    # sample and concatenate the IQ pulses
    samples = np.hstack([modulate_iq(i) for i in flushed_iq_pulses])

    # normalize the real and imaginary waveform components
    def normalize(samples: np.ndarray) -> tuple[np.ndarray, float]:
        """Normalize real-valued samples to [-1, 1]."""
        scale = np.max(np.abs(samples))
        # avoid division by zero
        if scale > 0:
            return samples / scale, scale
        return samples, 1.0

    samples.real, scale_i = normalize(samples.real)
    samples.imag, scale_q = normalize(samples.imag)
    return IQPulse(
        duration=len(samples),
        wave_i=Samples(samples.real),
        wave_q=Samples(samples.imag),
        scale_i=scale_i,
        scale_q=scale_q,
        phase=fused_phase,
        phase_increment=np.sum(phase_increments),
        modulation_frequency=0.0,  # modulate_iq takes care of this
    )



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Projective measurement in the Z basis."""

from __future__ import annotations

from collections.abc import Iterable
from copy import copy, deepcopy
from dataclasses import replace
import functools
from typing import TYPE_CHECKING

import numpy as np

from exa.common.data.parameter import CollectionType, DataType, Parameter, Setting
from iqm.pulse.gate_implementation import (
    PROBE_LINES_LOCUS_MAPPING,
    SINGLE_COMPONENTS_WITH_READOUT_LOCUS_MAPPING,
    CompositeGate,
    CustomIQWaveforms,
    Locus,
    OILCalibrationData,
)
from iqm.pulse.playlist import Schedule, Segment
from iqm.pulse.playlist.channel import ProbeChannelProperties
from iqm.pulse.playlist.instructions import (
    AcquisitionMethod,
    Block,
    ComplexIntegration,
    IQPulse,
    MultiplexedIQPulse,
    ReadoutTrigger,
    ThresholdStateDiscrimination,
    TimeTrace,
)
from iqm.pulse.playlist.waveforms import Constant, Samples
from iqm.pulse.timebox import MultiplexedProbeTimeBox, ProbeTimeBoxes, SchedulingStrategy, TimeBox

if TYPE_CHECKING:  # pragma: no cover
    from iqm.pulse.builder import ScheduleBuilder
    from iqm.pulse.quantum_ops import QuantumOp

DEFAULT_INTEGRATION_KEY = "readout.result"
DEFAULT_TIME_TRACE_KEY = "readout.time_trace"
FEEDBACK_KEY = "feedback"
TIMING_TOLERANCE = 1e-12


class Measure_CustomWaveforms(CustomIQWaveforms):
    """Base class for implementing dispersive measurement operations with custom probe pulse waveforms.

    You may define a measurement implementation that uses the :class:`.Waveform`
    instances ``Something`` and ``SomethingElse`` as the probe pulse waveforms in the
    I and Q channels as follows:
    ``class MyGate(Measure_CustomWaveforms, i_wave=Something, q_wave=SomethingElse)``.

    The ``measure`` operation is factorizable, and its :attr:`arity` is 0, which together mean that it can operate
    on loci of any length, but is calibrated only on single component loci. When the gate is constructed in the
    ``len(locus) > 1`` case (e.g. ``builder.get_implementation('measure', ('QB1', 'QB2', 'QB3'))()``), the resulting
    :class:`.TimeBox` is constructed from the calibrated single-component implementations.

    For each measured component, the readout :class:`.IQPulse` will be modulated with the
    intermediate frequency (IF), computed as the difference between the readout
    frequency of that component and the probe line center frequency, and offset in phase
    by the readout phase of the component.

    The measurement is implemented using a :class:`.ReadoutTrigger` instruction, with a duration set by the
    requirements of the acquisition(s). Note that this is typically different from
    ``gates.measure.constant.{locus}.duration``, which is the probe pulse duration.
    """

    root_parameters = {
        "duration": Parameter("", "Readout pulse duration", "s"),
        "frequency": Parameter("", "Readout pulse frequency", "Hz"),
        "phase": Parameter("", "Readout pulse phase", "rad"),
        "amplitude_i": Parameter("", "Readout channel I amplitude", ""),
        # TODO do we really need these defaults? are they used anywhere?
        "amplitude_q": Setting(Parameter("", "Readout channel Q amplitude", ""), 0.0),
        "integration_length": Parameter("", "Integration length", "s"),
        "integration_weights_I": Setting(
            Parameter("", "Integration weights for channel I", "", collection_type=CollectionType.NDARRAY),
            np.array([]),
        ),
        "integration_weights_Q": Setting(
            Parameter("", "Integration weights for channel Q", "", collection_type=CollectionType.NDARRAY),
            np.array([]),
        ),
        "integration_threshold": Parameter("", "Integration threshold", ""),
        "acquisition_type": Setting(Parameter("", "Acquisition type", "", data_type=DataType.STRING), "threshold"),
        "acquisition_delay": Parameter("", "Acquisition delay", "s"),
    }

    def __init__(
        self, parent: QuantumOp, name: str, locus: Locus, calibration_data: OILCalibrationData, builder: ScheduleBuilder
    ):
        super().__init__(parent, name, locus, calibration_data, builder)

        self._multiplexed_timeboxes: dict[tuple[str, str, bool], TimeBox] = {}
        """Cache for :meth:`probe_timebox`."""
        self._time_traces: dict[tuple[str, float | None, float | None, str], TimeBox] = {}
        """Cache for :meth:`time_trace`."""
        self._neighborhood_components: set[str] = set(self.locus) | set(
            self.builder.chip_topology.component_to_probe_line[q] for q in self.locus
        )

        if len(locus) == 1:
            # prepare the single-component measurement
            probe_line: ProbeChannelProperties = builder.channels[  # type: ignore[assignment]
                builder.get_probe_channel(locus[0])
            ]
            # readout duration is determined by the acquisition, probe pulses are truncated to fit this window
            self._duration = (
                probe_line.duration_to_int_samples(
                    probe_line.round_duration_to_granularity(
                        calibration_data["acquisition_delay"] + calibration_data["integration_length"]
                    )
                )
                + probe_line.integration_stop_dead_time
            )
            self._probe_offset = probe_line.integration_start_dead_time
            # "duration" is only used by the probe pulse
            waveform_params = self.convert_calibration_data(
                calibration_data,
                {k: v for k, v in self.parameters.items() if k not in self.root_parameters},
                probe_line,
            )
            # unconverted cal data that corresponds to a root param (not duration)
            root_params = {k: v for k, v in calibration_data.items() if k in self.root_parameters and k != "duration"}
            # do some conversions TODO are these consistent?
            root_params["integration_length"] = probe_line.duration_to_int_samples(root_params["integration_length"])
            root_params["acquisition_delay"] = round(probe_line.duration_to_samples(root_params["acquisition_delay"]))

            if_freq = (calibration_data["frequency"] - probe_line.center_frequency) / probe_line.sample_rate

            self._probe_instruction, self._acquisition_method = self._build_instructions(
                waveform_params, root_params, if_freq
            )

    def _build_instructions(
        self, waveform_params: OILCalibrationData, root_params: OILCalibrationData, if_freq: float
    ) -> tuple[IQPulse, AcquisitionMethod]:
        """Builds a probe pulse and an acquisition method using the calibration data.

        Subclasses may override this method if needed.
        """
        if self.dependent_waves:
            wave_i = self.wave_i(**waveform_params)
            wave_q = self.wave_q(**waveform_params)
        else:
            wave_i = self.wave_i(**waveform_params["i"])
            wave_q = self.wave_q(**waveform_params["q"])

        probe_pulse = IQPulse(
            duration=waveform_params["n_samples"],
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=root_params["amplitude_i"],
            scale_q=root_params["amplitude_q"],
            phase=root_params["phase"],
            modulation_frequency=if_freq,
        )

        integration_length = root_params["integration_length"]
        weights_i = root_params.get("integration_weights_I")
        weights_q = root_params.get("integration_weights_Q")
        if weights_i is not None and weights_i.size and weights_q is not None and weights_q.size:
            # TODO: the weights should be in the params, so we should not need to check that
            # make sure everything indeed works like this
            if not integration_length == weights_i.size == weights_q.size:
                raise ValueError(
                    "Integration length does not match with the provided integration weight lengths. "
                    f"For {self.locus}: the integration length is {integration_length} samples, "
                    f" the I weights vector length is {weights_i.size}, and the Q weights vector length"
                    f" is {weights_q.size}."
                )
            weights = IQPulse(
                duration=integration_length,
                wave_i=Samples(weights_i),
                wave_q=Samples(weights_q),
                scale_i=1.0,
                scale_q=1.0,
                phase=0.0,
                modulation_frequency=if_freq,  # TODO: should be fixed to -if_freq in Programmable RO Phase2?
            )
        else:
            const = Constant(integration_length)
            weights = IQPulse(
                duration=integration_length,
                wave_i=const,
                wave_q=const,
                scale_i=1.0,
                scale_q=0.0,
                phase=0.0,
                modulation_frequency=if_freq,  # TODO: should be fixed to -if_freq in Programmable RO Phase2?
            )

        acquisition_type = root_params.get("acquisition_type", self.root_parameters["acquisition_type"].value)  # type: ignore[union-attr]
        acquisition_label = "TO_BE_REPLACED"
        op_and_implementation = f"{self.parent.name}.{self.name}"
        if acquisition_type == "complex":
            acquisition_method = ComplexIntegration(
                label=acquisition_label,
                delay_samples=root_params["acquisition_delay"],
                weights=weights,
                implementation=op_and_implementation,
            )
        elif acquisition_type == "threshold":
            acquisition_method = ThresholdStateDiscrimination(
                label=acquisition_label,
                delay_samples=root_params["acquisition_delay"],
                weights=weights,
                threshold=root_params["integration_threshold"],
                implementation=op_and_implementation,
            )
        else:
            raise ValueError(f"Unknown acquisition type {acquisition_type}")

        return probe_pulse, acquisition_method

    def probe_timebox(self, key: str = "", feedback_key: str = "", do_acquisition: bool = True, **kwargs) -> TimeBox:
        """Returns a "naked" probe timebox that supports convenient multiplexing through
        ``MultiplexedProbeTimeBox.__add__``.

        This method can be used if the user wants to control the multiplexing explicitly. With two
        ``MultiplexedProbeTimeBox``es ``A`` and ``B`` the result ``A + B`` has all the ``ReadoutTrigger`` instructions
        on each probe channel of ``A`` and ``B`` multiplexed together and played simultaneously.

        Args:
            key: The readout results generated on this trigger will be assigned to
                ``f"{qubit}__{key}"``, where ``qubit`` goes over the component names in ``self.locus``. If empty,
                the key `"readout.result"` will be used to maintain backwards compatibility.
            feedback_key: The signals generated by this measure operation are routed using this key for
                fast feedback purposes. See :meth:`__call__`.
            do_acquisition: if False, no acquisitions are added.

        Returns:
            MultiplexedProbeTimeBox containing the ReadoutTrigger instruction.

        """
        args = (key, feedback_key, do_acquisition)
        # additional caching for probe timeboxes due to the fact that both ._call and .time_trace use this
        if args not in self._multiplexed_timeboxes:
            if len(self.locus) == 1:
                label_key = key or DEFAULT_INTEGRATION_KEY
                replacements = {"label": f"{self.locus[0]}__{label_key}"}
                if feedback_key and isinstance(self._acquisition_method, ThresholdStateDiscrimination):
                    # TODO: use the actual ``feedback_key`` when AWGs support multiple feedback labels
                    replacements["feedback_signal_label"] = f"{self.locus[0]}__{FEEDBACK_KEY}"
                acquisitions = (replace(self._acquisition_method, **replacements),) if do_acquisition else ()  # type: ignore[arg-type]
                multiplexed_iq = MultiplexedIQPulse(
                    duration=self._probe_instruction.duration + self._probe_offset,
                    entries=((self._probe_instruction, self._probe_offset),),
                )
                readout_trigger = ReadoutTrigger(
                    duration=self._duration, probe_pulse=multiplexed_iq, acquisitions=acquisitions
                )
                probe_channel = self.builder.get_probe_channel(self.locus[0])
                try:
                    drive_channel = self.builder.get_drive_channel(self.locus[0])
                except KeyError:
                    drive_channel = ""

                if drive_channel:
                    # drive channel must be blocked, to prevent DD insertion while measurement is taking place
                    # unfortunately we must allow for different channel sample rates because of UHFQA
                    channels = self.builder.channels
                    drive_channel_props = channels[drive_channel]
                    rt_duration_in_seconds = channels[probe_channel].duration_to_seconds(readout_trigger.duration)
                    block_duration = drive_channel_props.duration_to_int_samples(
                        drive_channel_props.round_duration_to_granularity(
                            rt_duration_in_seconds, round_up=True, force_min_duration=True
                        )
                    )
                    probe_timebox = MultiplexedProbeTimeBox.from_readout_trigger(
                        readout_trigger=readout_trigger,
                        probe_channel=probe_channel,
                        locus_components=self.locus,
                        label=f"{self.__class__.__name__} on {self.locus}",
                        block_channels=[drive_channel],
                        block_duration=block_duration,
                    )
                else:
                    probe_timebox = MultiplexedProbeTimeBox.from_readout_trigger(
                        readout_trigger=readout_trigger,
                        probe_channel=probe_channel,
                        locus_components=self.locus,
                        label=f"{self.__class__.__name__} on {self.locus}",
                    )
            else:
                probe_timeboxes = [
                    self.sub_implementations[c].probe_timebox(key, feedback_key, do_acquisition)  # type: ignore[attr-defined]
                    for c in self.locus
                ]
                probe_timebox = functools.reduce(lambda x, y: x + y, probe_timeboxes)
            if isinstance(probe_timebox, TimeBox):  # FIXME: not needed once the measure implementations are cleaned up
                probe_timebox.neighborhood_components[0] = copy(self._neighborhood_components)
                if feedback_key:
                    # Block all the virtual channels from the probes involved in self.locus as we cannot know what AWG
                    # might be listening to the sent bits. NOTE: No Waits are added, the channels are just blocked in
                    # scheduling, so the impact to performance is negligible
                    probelines = {self.builder.chip_topology.component_to_probe_line[q] for q in self.locus}
                    for probe in probelines:
                        probe_timebox.neighborhood_components[0].update(
                            set(self.builder.get_virtual_feedback_channels(probe))
                        )
            self._multiplexed_timeboxes[args] = probe_timebox
        return self._multiplexed_timeboxes[args]

    def _call(self, key: str = "", feedback_key: str = "") -> TimeBox:
        """Returns a TimeBox containing the multiplexed simultaneous measurement.

        If ``len(self.locus) == 1``, the TimeBox contains the measurement for just that component, otherwise
        the measurements of components that belong to the same probeline are multiplexed together.

        The returned :class:`.TimeBox` instances behave like any other TimeBox in scheduling and circuit
        generation. With measurement TimeBoxes ``A`` and ``B`` the result ``A + B`` first plays the ``ReadoutTrigger``
        instructions of ``A`` and only then those of ``B`` in each probe channel. If the multiplexing features of
        :class:`.MultiplexedProbeTimeBox` are needed, the method :meth:`probe_timebox` can be used.

        In scheduling, the returned TimeBox blocks the locus components and the probe
        lines they are associated with.

        Args:
            key: Readout results generated on this trigger will be assigned to the acquisition labels
                ``f"{qubit}__{key}"``, where ``qubit`` goes over the component names in ``self.locus``.
                If empty, the key ``"readout.result"`` will be used to maintain backwards compatibility.
            feedback_key: If the readout mode is "threshold", the results generated by this ``measure`` operation
                are routed using the label ``f"{qubit}__{feedback_key}"`` for fast feedback purposes.
                The signals are picked up by :class:`.ConditionalInstruction`s that have the same label.
                The default value ``""`` means the signal is not routed anywhere. TODO: currently the HW does not
                support multiple feedback keys per drive channel, so the actual key used will be ``FEEDBACK_KEY``
                whenever any non-empty key is inputted. When the HW is improved, the actual key the user inputs
                should be passed.

        Returns:
            TimeBox containing the :class:`.ReadoutTrigger` instruction.

        """
        final_box = TimeBox.composite(
            [self.probe_timebox(key=key, feedback_key=feedback_key)], label=f"Readout on {self.locus}"
        )
        final_box.neighborhood_components[0] = final_box.children[0].neighborhood_components[0]
        return final_box

    def _get_probe_timebox_for_time_trace(self, key: str = "", feedback_key: str = "") -> TimeBox:
        """Utility method that can be overridden in subclasses if they have a return type `.probe_pulse`."""
        # FIXME: not needed once we align the return types of all these measure gates
        return self.probe_timebox(key=key, feedback_key=feedback_key)

    def time_trace(
        self,
        key: str = "",
        acquisition_delay: float | None = None,
        acquisition_duration: float | None = None,
        feedback_key: str = "",
    ) -> TimeBox:
        """Returns a multiplexed simultaneous measurement with an additional time trace acquisition.

        The returned ``TimeBox`` is the same as the one returned by :meth:`__call__` except the time trace
        acquisition is appended to the acquisitions of each probe line's ``ReadoutTrigger`` instruction.

        Args:
            key: Readout results generated on this trigger will be used to assigned to
                ``f"{qubit}__{key}"``, where ``qubit`` goes over the component names in ``self.locus``, whereas
                the recorded time traces will be assigned to ``f"{probe_line}__{key}"`` where
                ``probe_line`` goes through all the probe lines associated with ``self.locus``.
                If empty, the key ``"readout.result"`` will be used for integrated results and the key
                ``"readout.time_trace"`` for the recorded time traces.
            acquisition_delay: optionally override the time trace acquisition delay with this value (given in
                seconds). Does not affect the acqusition delays of the integrated measurements.
            acquisition_duration: optionally override the time trace acquisition duration with this value (given in
                seconds). Does not affect the integration lengths of the integrated measurements.
            feedback_key: The signals generated by the integration are routed using this label, prefixed by
                the component. See :meth:`__call__`.

        Returns:
            TimeBox containing the ReadoutTrigger instruction.

        """
        args = (key, acquisition_delay, acquisition_duration, feedback_key)
        # additional caching for time traces since the acquisitions differ from the ones in _call
        if args not in self._time_traces:
            probe_timebox = deepcopy(self._get_probe_timebox_for_time_trace(key, feedback_key))
            for probe_channel, segment in probe_timebox.atom.items():  # type: ignore[union-attr]
                readout_trigger = None
                for inst in segment:
                    if isinstance(inst, ReadoutTrigger):
                        readout_trigger = inst
                        break
                # TODO instead of editing the probe_timebox output contents, we should make the function itself do this
                # so we would not need to blindly search through the channels
                if readout_trigger is None:
                    continue

                probe_line = self.builder.channels[probe_channel]
                probe_name = self.builder._channel_to_component[probe_channel]

                if acquisition_delay is not None:
                    delay_samples = probe_line.duration_to_int_samples(acquisition_delay)
                else:
                    delay_samples = min(acq.delay_samples for acq in readout_trigger.acquisitions)
                if acquisition_duration is not None:
                    duration_samples = probe_line.duration_to_int_samples(acquisition_duration)
                else:
                    duration_samples = max(
                        acq.weights.duration + acq.delay_samples - delay_samples  # type: ignore[attr-defined]
                        for acq in readout_trigger.acquisitions
                    )
                label_key = key or DEFAULT_TIME_TRACE_KEY
                time_trace = TimeTrace(
                    label=f"{probe_name}__{label_key}",
                    delay_samples=delay_samples,
                    duration_samples=duration_samples,
                    implementation=f"{self.parent.name}.{self.name}",
                )
                trigger_with_trace = replace(readout_trigger, acquisitions=readout_trigger.acquisitions + (time_trace,))
                segment._instructions[0] = trigger_with_trace

            final_box = TimeBox.composite([probe_timebox])
            final_box.neighborhood_components[0] = probe_timebox.neighborhood_components[0]
            self._time_traces[args] = final_box
        return self._time_traces[args]

    def duration_in_seconds(self) -> float:
        probe_timebox = self.probe_timebox()
        readout_schedule = probe_timebox.atom
        return readout_schedule.duration_in_seconds(self.builder.channels)  # type: ignore[union-attr]

    @classmethod
    def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
        return SINGLE_COMPONENTS_WITH_READOUT_LOCUS_MAPPING


class Measure_Constant(Measure_CustomWaveforms, wave_i=Constant, wave_q=Constant):  # type: ignore[call-arg]
    """Implementation of a single-qubit projective, dispersive measurement in the Z basis.

    Uses a constant probe pulse.
    """


class Measure_Constant_Qnd(Measure_CustomWaveforms, wave_i=Constant, wave_q=Constant):  # type:ignore[call-arg]
    """Implementation of a single-qubit projective, non quantum demolition, dispersive
    measurements in the Z basis.

    Uses a constant probe pulse.
    """


class ProbePulse_CustomWaveforms(CustomIQWaveforms):
    """Base class for implementing a probe line measurement pulse with custom waveforms in the I and Q channels.

    With given :class:`.Waveform` waveform definitions ``Something`` and ``SomethingElse``,
    you may define a measurement implementation that uses them as follows:
    ``class MyGate(ProbePulse_CustomWaveforms, i_wave=Something, q_wave=SomethingElse)``.
    The measurement :class:`.IQPulse` instruction will not be automatically modulated
    by any frequency, so any modulations should be included in the I and Q waveforms themselves.

    Due to device limitations this implementation also has to integrate the readout signal
    (using arbitrary weights), even though it does not make much sense.

    Contrary to the ``Measure_CustomWaveforms`` class, this implementation acts on proble lines directly (i.e. its
    ``locus`` is a single probe line).
    """

    root_parameters = {
        "duration": Parameter("", "Readout pulse duration", "s"),
        "phase": Parameter("", "Readout pulse phase", "rad"),
        "amplitude_i": Parameter("", "Readout channel I amplitude", ""),
        "amplitude_q": Parameter("", "Readout channel Q amplitude", ""),
        "integration_length": Parameter("", "Integration length", "s"),
        "acquisition_delay": Parameter("", "Acquisition delay", "s"),
    }

    def __init__(
        self, parent: QuantumOp, name: str, locus: Locus, calibration_data: OILCalibrationData, builder: ScheduleBuilder
    ):
        super().__init__(parent, name, locus, calibration_data, builder)
        self._probe_line: ProbeChannelProperties = builder.channels[  # type: ignore[assignment]
            builder.component_channels[locus[0]]["readout"]
        ]
        self._duration = (
            self._probe_line.duration_to_int_samples(
                self._probe_line.round_duration_to_granularity(
                    calibration_data["acquisition_delay"] + calibration_data["integration_length"]
                )
            )
            + self._probe_line.integration_stop_dead_time
        )

        waveform_params = self.convert_calibration_data(
            calibration_data,
            {k: v for k, v in self.parameters.items() if k not in self.root_parameters},
            self._probe_line,
        )
        root_params = {k: v for k, v in calibration_data.items() if k in self.root_parameters and k != "duration"}
        probe_instruction, acquisitions = self._build_instructions(waveform_params, root_params)
        self._probe_instruction = probe_instruction
        self._acquisitions = acquisitions

    def _build_instructions(
        self, waveform_params: OILCalibrationData, root_params: OILCalibrationData
    ) -> tuple[IQPulse, tuple[AcquisitionMethod, AcquisitionMethod]]:
        """Builds a probe pulse and acquisition methods using the calibration data.

        Subclasses may override this method if needed.
        """
        if self.dependent_waves:
            wave_i = self.wave_i(**waveform_params)
            wave_q = self.wave_q(**waveform_params)
        else:
            wave_i = self.wave_i(**waveform_params["i"])
            wave_q = self.wave_q(**waveform_params["q"])

        probe_pulse = IQPulse(
            duration=waveform_params["n_samples"],
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=root_params["amplitude_i"],
            scale_q=root_params["amplitude_q"],
            phase=root_params["phase"],
            modulation_frequency=0,
        )

        integration_length = self._probe_line.duration_to_int_samples(root_params["integration_length"])
        acquisition_delay = round(self._probe_line.duration_to_samples(root_params["acquisition_delay"]))
        time_trace_label = "TO_BE_REPLACED"
        time_trace_acquisition = TimeTrace(
            label=time_trace_label,
            delay_samples=acquisition_delay,
            duration_samples=integration_length,
            implementation=f"{self.parent.name}.{self.name}",
        )

        # TODO: due to device limitations, we need to integrate always, even though it does not make much sense here
        const = Constant(integration_length)
        weights = IQPulse(
            duration=integration_length,
            wave_i=const,
            wave_q=const,
            scale_i=1.0,
            scale_q=0.0,
            phase=0.0,
            modulation_frequency=0,
        )
        integration_label = "dummy__integration"
        integration_acquisition = ComplexIntegration(
            label=integration_label,
            delay_samples=acquisition_delay,
            weights=weights,
            implementation=f"{self.parent.name}.{self.name}",
        )
        return probe_pulse, (integration_acquisition, time_trace_acquisition)

    def _call(self, key: str = "") -> TimeBox:
        """Returns a ``TimeBox`` containing the probe pulse measurement.

        In scheduling, the returned ``TimeBox`` blocks only the probe line (``self.locus[0]``).

        Args:
            key: The time trace results generated on this trigger will be used to assigned to
                ``f"{probe_line}__{key}"``, where ``probe_line`` is the one that handles ``self.locus[0]``. If empty,
                the key `"readout.time_trace"` is used.

        Returns:
            TimeBox containing the ReadoutTrigger instruction.

        """
        probe_channel = self.builder.component_channels[self.locus[0]]["readout"]
        multiplexed_iq = MultiplexedIQPulse(
            duration=self._probe_instruction.duration + self._probe_line.integration_start_dead_time,
            entries=((self._probe_instruction, self._probe_line.integration_start_dead_time),),
        )
        label_key = key or DEFAULT_TIME_TRACE_KEY
        acquisition_label = f"{self.locus[0]}__{label_key}"
        acquisitions = (self._acquisitions[0], replace(self._acquisitions[1], label=acquisition_label))
        readout_trigger = ReadoutTrigger(duration=self._duration, probe_pulse=multiplexed_iq, acquisitions=acquisitions)
        probe_timebox = MultiplexedProbeTimeBox.from_readout_trigger(
            readout_trigger=readout_trigger,
            probe_channel=probe_channel,
            locus_components=(self.locus[0],),
            label=f"{self.__class__.__name__} on {self.locus}",
        )
        final_box = TimeBox.composite([probe_timebox], label=probe_timebox.label)
        final_box.neighborhood_components[0] = {self.locus[0]}
        return final_box

    def duration_in_seconds(self) -> float:
        probe_timebox = self().children[0]  # type: ignore[union-attr]
        probe_channel = self.builder.component_channels[self.locus[0]]["readout"]
        return self.builder.channels[probe_channel].duration_to_seconds(probe_timebox.atom.duration)  # type: ignore[union-attr]

    @classmethod
    def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
        return PROBE_LINES_LOCUS_MAPPING


class ProbePulse_CustomWaveforms_noIntegration(CustomIQWaveforms):
    """Base class for implementing a probe line probe pulse with custom waveforms in the I and Q channels without
    any integration.

    Similar to the :class:`ProbePulse_CustomWaveforms` except that signal integration is removed.
    """

    root_parameters = {
        "frequency": Parameter("", "Readout pulse frequency", "Hz"),
        "duration": Parameter("", "Readout pulse duration", "s"),
        "phase": Parameter("", "Readout pulse phase", "rad"),
        "amplitude_i": Parameter("", "Readout channel I amplitude", ""),
        "amplitude_q": Parameter("", "Readout channel Q amplitude", ""),
    }

    def __init__(
        self, parent: QuantumOp, name: str, locus: Locus, calibration_data: OILCalibrationData, builder: ScheduleBuilder
    ):
        super().__init__(parent, name, locus, calibration_data, builder)

        self._multiplexed_timeboxes: dict[tuple[str, str, bool], MultiplexedProbeTimeBox] = {}
        """Cache for :meth:`probe_timebox`."""

        if len(locus) == 1:  # factorizable gates only need calibration on 1-loci
            self._probe_line: ProbeChannelProperties = builder.channels[  # type: ignore[assignment]
                builder.get_probe_channel(locus[0])
            ]
            c_freq = self._probe_line.center_frequency
            if_freq = (calibration_data["frequency"] - c_freq) / self._probe_line.sample_rate
            self._duration = (
                self._probe_line.duration_to_int_samples(
                    self._probe_line.round_duration_to_granularity(calibration_data["duration"])
                )
                + self._probe_line.instruction_duration_granularity
            )

            waveform_params = self.convert_calibration_data(
                calibration_data,
                {k: v for k, v in self.parameters.items() if k not in self.root_parameters or k == "duration"},
                self._probe_line,
            )
            root_params = {k: v for k, v in calibration_data.items() if k in self.root_parameters and k != "duration"}

            probe_instruction = self._build_instructions(waveform_params, root_params, if_freq)
            self._probe_instruction = probe_instruction
            self._prio_calibration: OILCalibrationData | None = None
        else:
            # we need to store the possible cal_data == priority calibration in order to propagate it to the factored
            # single-component measure calls in :meth:`probe_timebox`
            self._prio_calibration = calibration_data or None

    def _build_instructions(
        self, waveform_params: OILCalibrationData, root_params: OILCalibrationData, if_freq: float
    ) -> IQPulse:
        """Builds a probe pulse and an acquisition method using the calibration data.

        Subclasses may override this method if needed.
        """
        if self.dependent_waves:
            wave_i = self.wave_i(**waveform_params)
            wave_q = self.wave_q(**waveform_params)
        else:
            wave_i = self.wave_i(**waveform_params["i"])
            wave_q = self.wave_q(**waveform_params["q"])

        probe_pulse = IQPulse(
            duration=waveform_params["n_samples"],
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=root_params["amplitude_i"],
            scale_q=root_params["amplitude_q"],
            phase=root_params["phase"],
            modulation_frequency=if_freq,
        )

        return probe_pulse

    def probe_timebox(
        self, key: str = "", feedback_key: str = "", do_acquisition: bool = False
    ) -> MultiplexedProbeTimeBox:
        """Returns a "naked" probe timebox that supports convenient multiplexing through
        ``MultiplexedProbeTimeBox.__add__``.

        This method can be used if the user wants to control the multiplexing explicitly. With two
        ``MultiplexedProbeTimeBox``es ``A`` and ``B`` the result ``A + B`` has all the ``ReadoutTrigger`` instructions
        on each probe channel of ``A`` and ``B`` multiplexed together and played simultaneously.

        Args:
            key: The readout results generated on this trigger will be assigned to
                ``f"{qubit}__{key}"``, where ``qubit`` goes over the component names in ``self.locus``. If empty,
                the key `"readout.result"` will be used to maintain backwards compatibility.
            feedback_key: The signals generated by this measure operation are routed using this key for
                fast feedback purposes. See :meth:`__call__`.
            do_acquisition: if False, no acquisitions are added.

        Returns:
            MultiplexedProbeTimeBox containing the ReadoutTrigger instruction.

        """
        args = (key, feedback_key, do_acquisition)
        # additional caching for probe timeboxes due to the fact that both ._call and .time_trace use this
        if args not in self._multiplexed_timeboxes:
            if len(self.locus) == 1:
                multiplexed_iq = MultiplexedIQPulse(
                    duration=self._probe_instruction.duration,
                    entries=((self._probe_instruction, 0),),
                )
                readout_trigger = ReadoutTrigger(
                    probe_pulse=multiplexed_iq,
                    acquisitions=(),
                    duration=self._duration,
                )
                probe_channel = self.builder.get_probe_channel(self.locus[0])
                try:
                    drive_channel = self.builder.get_drive_channel(self.locus[0])
                except KeyError:
                    drive_channel = ""

                if drive_channel:
                    # drive channel must be blocked, to prevent DD insertion while measurement is taking place
                    # unfortunately we must allow for different channel sample rates because of UHFQA
                    channels = self.builder.channels
                    drive_channel_props = channels[drive_channel]
                    rt_duration_in_seconds = channels[probe_channel].duration_to_seconds(readout_trigger.duration)
                    block_duration = drive_channel_props.duration_to_int_samples(
                        drive_channel_props.round_duration_to_granularity(
                            rt_duration_in_seconds, round_up=True, force_min_duration=True
                        )
                    )
                    probe_timebox = MultiplexedProbeTimeBox.from_readout_trigger(
                        readout_trigger=readout_trigger,
                        probe_channel=probe_channel,
                        locus_components=self.locus,
                        label=f"{self.__class__.__name__} on {self.locus}",
                        block_channels=[drive_channel],
                        block_duration=block_duration,
                    )
                else:
                    probe_timebox = MultiplexedProbeTimeBox.from_readout_trigger(
                        readout_trigger=readout_trigger,
                        probe_channel=probe_channel,
                        locus_components=self.locus,
                        label=f"{self.__class__.__name__} on {self.locus}",
                    )
            else:
                probe_timeboxes = [
                    self.builder.get_implementation(  # type: ignore[attr-defined]
                        self.parent.name, (c,), impl_name=self.name, priority_calibration=self._prio_calibration
                    ).probe_timebox(key, feedback_key, do_acquisition)
                    for c in self.locus
                ]
                probe_timebox = functools.reduce(lambda x, y: x + y, probe_timeboxes)
            probe_timebox.neighborhood_components[0] = set(
                self.locus + (self.builder.chip_topology.component_to_probe_line[self.locus[0]],)
            )

            self._multiplexed_timeboxes[args] = probe_timebox
        return self._multiplexed_timeboxes[args]

    def _call(self, key: str = "", feedback_key: str = "") -> TimeBox:
        """Returns a TimeBox containing the multiplexed simultaneous measurement.

        If ``len(self.locus) == 1``, the TimeBox contains the measurement for just that component, otherwise
        the measurements of components that belong to the same probeline are multiplexed together.

        The returned :class:`.TimeBox` instances behave like any other TimeBox in scheduling and circuit
        generation. With measurement TimeBoxes ``A`` and ``B`` the result ``A + B`` first plays the ``ReadoutTrigger``
        instructions of ``A`` and only then those of ``B`` in each probe channel. If the multiplexing features of
        :class:`.MultiplexedProbeTimeBox` are needed, the method :meth:`probe_timebox` can be used.

        In scheduling, the returned TimeBox blocks the locus components and the probe
        lines they are associated with.

        Args:
            key: Readout results generated on this trigger will be assigned to the acquisition labels
                ``f"{qubit}__{key}"``, where ``qubit`` goes over the component names in ``self.locus``.
                If empty, the key ``"readout.result"`` will be used to maintain backwards compatibility.
            feedback_key: If the readout mode is "threshold", the results generated by this ``measure`` operation
                are routed using the label ``f"{qubit}__{feedback_key}"`` for fast feedback purposes.
                The signals are picked up by :class:`.ConditionalInstruction`s that have the same label.
                The default value ``""`` means the signal is not routed anywhere. TODO: currently the HW does not
                support multiple feedback keys per drive channel, so the actual key used will be ``FEEDBACK_KEY``
                whenever any non-empty key is inputted. When the HW is improved, the actual key the user inputs
                should be passed.

        Returns:
            TimeBox containing the :class:`.ReadoutTrigger` instruction.

        """
        final_box = TimeBox.composite(
            [self.probe_timebox(key=key, feedback_key=feedback_key)], label=f"Readout on {self.locus}"
        )
        final_box.neighborhood_components[0] = final_box.children[0].neighborhood_components[0]
        return final_box

    def duration_in_seconds(self) -> float:
        probe_timebox = self.probe_timebox()
        readout_schedule = probe_timebox.atom
        return readout_schedule.duration_in_seconds(self.builder.channels)  # type: ignore[union-attr]

    @classmethod
    def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
        return SINGLE_COMPONENTS_WITH_READOUT_LOCUS_MAPPING


class Probe_Constant(ProbePulse_CustomWaveforms_noIntegration, wave_i=Constant, wave_q=Constant):  # type: ignore[call-arg]
    """Implementation of a single-qubit projective, dispersive measurement in the Z basis.

    Uses a constant probe pulse.
    """


class ShelvedMeasureTimeBox(TimeBox):
    """TimeBox representing a shelved measurement (ReadoutTrigger sandwiched between two PRX_12 operations).

    ShelvedMeasureTimeBox is a composite TimeBox containing two children:
    * first one being the first PRX_12 operation for the locus components of the measure
    * second one being the ReadoutTrigger (MultiplexedProbeTimeBox) that includes the second PRX_12 operation.

    Multiplexing is achieved so that ShelvedMeasureTimeBoxes support ``__add__`` and ``__radd__`` operations with other
    boxes of the same type and MultiplexedProbeTimeBoxes. The multiplexing operation is defined such that the
    initial PRX_12 boxes are added together (in case one of the multiplexed boxes is a MultiplexedProbeTimeBoxes, the
    initial PRX_12 is considered empty), and the probe boxes are multiplexed together via the logic defined in
    ``MultiplexedProbeTimeBoxes.__add__``. This behaviour results in the correct timings of the associated pulses
    after the multiplexing.

    """

    def __post_init__(self):
        if len(self.children) != 2:
            raise ValueError(
                "ShelvedMeasureTimeBox must have exactly two children: the first one corresponding to the "
                "initial prx_12 operations, and the second one to the ReadoutTrigger and the final prx_12"
            )
        if self.children[1].atom is None or not isinstance(self.children[1], MultiplexedProbeTimeBox):
            raise ValueError("The second child must be an atomic MultiplexedProbeTimeBox.")

    @property
    def prx_12_box(self) -> TimeBox:
        return self.children[0]

    @property
    def trigger_box(self) -> TimeBox:
        return self.children[1]

    def __add__(self, other: TimeBox | Iterable[TimeBox]) -> TimeBox:
        """Add the initial PRX_12 boxes together via the ``TimeBox``"""
        if isinstance(other, (ShelvedMeasureTimeBox, MultiplexedProbeTimeBox)):
            if isinstance(other, ShelvedMeasureTimeBox):
                prx_12_box = self.prx_12_box + other.prx_12_box
                trigger_box = self.trigger_box + other.trigger_box
            else:
                prx_12_box = self.prx_12_box
                trigger_box = self.trigger_box + other
            locus_components = self.locus_components.union(other.locus_components)
            multiplexed = ShelvedMeasureTimeBox(
                label=f"Shelved measure on {locus_components}",
                locus_components=locus_components,
                atom=None,
                children=(prx_12_box, trigger_box),
                scheduling=self.scheduling,
                scheduling_algorithm=self.scheduling_algorithm,
            )
            # neighborhood components by the trigger_box
            multiplexed.neighborhood_components[0] = trigger_box.neighborhood_components[0]
            return multiplexed
        return super().__add__(other)

    def __radd__(self, other: TimeBox | Iterable[TimeBox]) -> TimeBox:
        if isinstance(other, MultiplexedProbeTimeBox):
            return self.__add__(other)  # this commutes
        return super().__radd__(other)


class Shelved_Measure_CustomWaveforms(Measure_CustomWaveforms, CompositeGate):
    """Base class for shelved readout.

    Shelved readout applies a ``prx_12(pi)`` gate before and after a standard dispersive readout on each qubit measured.
    The first ``prx_12(pi)`` swaps the amplitudes of the |1> and |2> states, and the second one swaps them back after
    the measurement has (roughly) collapsed the state. If the discriminator of the readout is calibrated such that
    the |0> state is on one side and the |1> and |2> states are on the other, the end result is equivalent to the
    standard readout operation but with the advantage that the population in the |2> state is less susceptible to
    :math:`T_1` decay during the readout than the population in the |1> state.

    """

    root_parameters = Measure_CustomWaveforms.root_parameters | {
        "second_prx_12_offset": Setting(
            Parameter(
                "second_prx_12_offset", "Offset of the second PRX_12 pulse from the end the ReadoutTrigger", unit="s"
            ),
            0.0,
        ),
        "do_prx_12": Setting(
            Parameter(
                "do_prx_12",
                "Whether to do the prx_12 flips in the measure operation",
                unit="",
                data_type=DataType.BOOLEAN,
            ),
            True,
        ),
    }
    registered_gates = ("prx_12",)

    def probe_timebox(self, key: str = "", feedback_key: str = "", do_acquisition: bool = True, **kwargs) -> TimeBox:  # type: ignore[union-attr, override]
        """Returns a "naked" probe timebox that supports convenient multiplexing through
        ``ShelvedMeasureTimeBox.__add__``.

        This method can be used if the user wants to control the multiplexing explicitly. Supports adding together
        boxes of type :class:`.ShelvedMeasureTimeBox` and/or :class:`.MultiplexedProbeTimeBox`. See
        :meth:`.ShelvedMeasureTimeBox.__add__` for more information on the logic.

        Args:
            key: The readout results generated on this trigger will be assigned to
                ``f"{qubit}__{key}"``, where ``qubit`` goes over the component names in ``self.locus``. If empty,
                the key `"readout.result"` will be used to maintain backwards compatibility.
            feedback_key: The signals generated by this measure operation are routed using this key for
                fast feedback purposes. See :meth:`__call__`.
            do_acquisition: if False, no acquisitions are added.

        Returns:
            ShelvedMeasureTimeBox containing the ReadoutTrigger instruction.

        """
        args = (key, feedback_key, do_acquisition)
        if args not in self._multiplexed_timeboxes:
            if len(self.locus) == 1:
                probe_timebox = super().probe_timebox(key, feedback_key, do_acquisition, **kwargs)
                shelved_box = probe_timebox
                prx_12_box = TimeBox.composite(
                    [self.build("prx_12", self.locus)(np.pi)], scheduling=SchedulingStrategy.ALAP
                )
                if self.calibration_data["do_prx_12"]:
                    shelved_box = probe_timebox + prx_12_box  # type: ignore[operator, assignment, override]
                # schedule the shelved box to get an atomic schedule
                shelved_atom = deepcopy(self.builder.resolve_timebox(shelved_box, neighborhood=0))
                offset = self.calibration_data["second_prx_12_offset"]
                if self.calibration_data["do_prx_12"] and abs(offset) > TIMING_TOLERANCE:
                    drive_channel_name = self.builder.get_drive_channel(self.locus[0])
                    drive_channel = self.builder.channels[drive_channel_name]
                    offset_sign = offset / abs(offset)
                    offset_in_samples = offset_sign * drive_channel.duration_to_int_samples(abs(offset))
                    trigger_block = shelved_atom[drive_channel_name][0]
                    block_with_offset = Block(trigger_block.duration + offset_in_samples)
                    shelved_atom[drive_channel_name]._instructions[0] = block_with_offset
                trigger_box = MultiplexedProbeTimeBox(
                    label=f"{self.__class__.__name__} on {self.locus}",
                    locus_components=probe_timebox.locus_components,
                    atom=shelved_atom,
                )
                trigger_box.neighborhood_components[0] = probe_timebox.neighborhood_components[0]
                pre_box = prx_12_box if self.calibration_data["do_prx_12"] else TimeBox.composite([])
                final_box = ShelvedMeasureTimeBox(
                    label=f"Shelved Measure on {self.locus}",
                    locus_components=set(self.locus),
                    atom=None,
                    children=(pre_box, trigger_box),
                )
                final_box.neighborhood_components[0] = probe_timebox.neighborhood_components[0]
            else:
                # NOTE: the super call can be a bit misleading; it is actually calling the `self.probe_timebox` of len 1
                # in this class inside, via the factorizable gate's sub_implementations
                final_box = super().probe_timebox(key, feedback_key)  # type: ignore[assignment]
            self._multiplexed_timeboxes[args] = final_box
        return self._multiplexed_timeboxes[args]

    def _get_probe_timebox_for_time_trace(self, key: str = "", feedback_key: str = "") -> TimeBox:
        """Utility method that can be overridden in subclasses if they have a return type `.probe_pulse`.

        The ``ShelvedMeasureTimeBox`` resulting from :meth:`.probe_timebox` is first scheduled to obtain an atomic
        ``MultiplexedProbeTimeBox`` which is wrapped into a TimeBox.
        """
        # FIXME: not needed once we align the return types of all these measure gates
        probe_timebox = self.probe_timebox(key=key, feedback_key=feedback_key)
        # resolve the box to get an atomic time_box.
        probe_schedule = self.builder.resolve_timebox(probe_timebox, neighborhood=0)
        atomic_probe_box = MultiplexedProbeTimeBox.atomic(
            probe_schedule,
            label=f"Time Trace atomic probe box of {self.__class__.__name__} on {self.locus}",
            locus_components=probe_timebox.locus_components,
        )
        atomic_probe_box.neighborhood_components[0] = probe_timebox.neighborhood_components[0]
        return atomic_probe_box


class Shelved_Measure_Constant(Shelved_Measure_CustomWaveforms, wave_i=Constant, wave_q=Constant):  # type:ignore[call-arg]
    """Implementation of a shelved readout.

    A measure gate implemented as a constant waveform is surrounded by two `prx_12` gates.
    """


class Fast_Measure_CustomWaveforms(Measure_CustomWaveforms):
    """Measure implementation that blocks locus qubits for a shorter duration than the probes.

    The locus qubits are blocked only for the physical probe pulse duration plus (calibratable) extra dead time that
    can be used to take into account e.g. ring down delay of waiting the readout resonator to empty itself. The probe
    channels are still blocked as in ``Measure_CustomWaveforms``, i.e. for the duration of
    ``acquisition_delay + integration_length + integration_dead_time``.
    """

    root_parameters = Measure_CustomWaveforms.root_parameters | {
        "locus_deadtime": Setting(
            Parameter("locus_deadtime", "Locus dead time after the probe pulse", unit="s"),
            0.0,
        ),
    }

    def probe_timebox(  # type: ignore[override]
        self, key: str = "", feedback_key: str = "", do_acquisition: bool = True, **kwargs
    ) -> ProbeTimeBoxes:
        """Otherwise the same as ``Measure_CustomWaveforms.probe_timebox``, but returns two TimeBoxes, the
        actual MultiplexedProbeTimeBox and the rest of the probe-blocking wait time in its own TimeBox. This
        allows the "tetris logic" in scheduling to block the locus qubits for a shorter duration.
        """
        args = (key, feedback_key, do_acquisition)
        if args not in self._multiplexed_timeboxes:
            if len(self.locus) == 1:
                probe = self.builder.chip_topology.component_to_probe_line[self.locus[0]]
                probe_channel = self.builder.get_probe_channel(self.locus[0])
                try:
                    drive_channel = self.builder.get_drive_channel(self.locus[0])
                except KeyError:
                    drive_channel = ""

                combined_probe_timebox = super().probe_timebox(
                    key, feedback_key, do_acquisition=do_acquisition, **kwargs
                )
                readout_trigger = combined_probe_timebox.atom[probe_channel][0]  # type: ignore[index]
                actual_probe_duration = readout_trigger.probe_pulse.duration

                # MultiplexedProbeTimeBox that has the minimum possible duration
                if self.calibration_data["locus_deadtime"] > TIMING_TOLERANCE:
                    deadtime = self.builder.channels[probe_channel].duration_to_int_samples(
                        self.calibration_data["locus_deadtime"]
                    )
                else:
                    deadtime = 0
                # Must be: ReadoutTrigger.duration > ReadoutTrigger.probe_pulse.duration so if they would match, we
                # need to add a minimum offset to make it hold, i.e. the smallest granularity allowed by the probe
                # channel
                probe_granularity = self.builder.channels[probe_channel].instruction_duration_granularity
                offset = max(deadtime, probe_granularity)

                truncated_readout_trigger = replace(readout_trigger, duration=actual_probe_duration + offset)
                physical_probe_box = MultiplexedProbeTimeBox.from_readout_trigger(
                    truncated_readout_trigger,
                    probe_channel,
                    locus_components=self.locus,
                    label=f"Physical probe box of {self.__class__.__name__} on {self.locus}",
                    block_channels=[drive_channel] if drive_channel else [],
                    block_duration=truncated_readout_trigger.duration if drive_channel else 0,
                )
                physical_probe_box.neighborhood_components[0] = combined_probe_timebox.neighborhood_components[0].copy()
                # extra Blocks (integration etc) for the probe channel
                extra_block_duration = max(combined_probe_timebox.atom.duration - truncated_readout_trigger.duration, 0)  # type:ignore[union-attr]
                virtual_extra_wait_box = TimeBox.atomic(
                    Schedule({probe_channel: Segment([Block(extra_block_duration)])}),
                    locus_components=[probe],
                    label=f"Virtual probe extra wait box of {self.__class__.__name__} on {self.locus}",
                )
                virtual_extra_wait_box.neighborhood_components[0] = {probe}
                final_boxes = ProbeTimeBoxes([physical_probe_box, virtual_extra_wait_box])
            else:
                # FIXME: all the return types in the measure impls need to be cleaned up
                final_boxes = super().probe_timebox(key, feedback_key, do_acquisition=do_acquisition, **kwargs)  # type:ignore[assignment]
            if feedback_key:
                probelines = {self.builder.chip_topology.component_to_probe_line[q] for q in self.locus}
                for probe in probelines:
                    final_boxes[1].neighborhood_components[0].update(
                        set(self.builder.get_virtual_feedback_channels(probe))
                    )
            self._multiplexed_timeboxes[args] = final_boxes  # type:ignore[assignment]
            return final_boxes  # type:ignore[return-value]
        return self._multiplexed_timeboxes[args]  # type:ignore[return-value]

    def _call(self, key: str = "", feedback_key: str = "") -> list[TimeBox]:  # type:ignore[override]
        """The same as ``Measure_CustomWaveforms._call``, i.e. wrap the "naked" multiplexable probe_timeboxes
        into a composite TimeBox.
        """
        probe_box, wait_box = self.probe_timebox(key, feedback_key, do_acquisition=True)
        final_box = TimeBox.composite([probe_box], label=f"Readout on {self.locus}")
        final_box.neighborhood_components[0] = final_box.children[0].neighborhood_components[0]
        return [final_box, wait_box]

    def _get_probe_timebox_for_time_trace(self, key: str = "", feedback_key: str = "") -> TimeBox:
        """Get the probe TimeBox for TimeTrace. This is just a single MultiplexedProbeTimeBox which has the full
        probe duration (including integration). The faster logic wrt. qubit blocking is not important in the context of
        TimeTraces.
        """
        # FIXME: not needed once we align the return types of all these measure gates
        # Then we can make this method also shorter for the locus qubits
        probe_boxes = self.probe_timebox(key, feedback_key, do_acquisition=True)
        total_box = TimeBox.composite(self.probe_timebox(key, feedback_key, do_acquisition=True))
        probe_schedule = self.builder.resolve_timebox(total_box, neighborhood=0)
        atomic_probe_box = MultiplexedProbeTimeBox.atomic(
            probe_schedule,
            label=f"Time Trace atomic probe box of {self.__class__.__name__} on {self.locus}",
            locus_components=probe_boxes[0].locus_components,
        )
        atomic_probe_box.neighborhood_components[0] = probe_boxes[0].neighborhood_components[0]
        return atomic_probe_box  # type:ignore[return-value]


class Fast_Measure_Constant(Fast_Measure_CustomWaveforms, wave_i=Constant, wave_q=Constant):  # type:ignore[call-arg]
    """Implementation of a faster measure with constant i and q waveforms.

    Does not block the drive and flux channels of the locus qubits during the integration, but just during the probe
    pulse and extra calibrated dead time after it.
    """



#  ********************************************************************************
#  Copyright (c) 2022-2025 IQM Finland Oy.
#  All rights reserved. Confidential and proprietary.
#
#  Distribution or reproduction of any information contained herein
#  is prohibited without IQM Finland Oy’s prior written permission.
#  ********************************************************************************
"""Representing quantum circuits as lists of CircuitOperations."""

from __future__ import annotations

from collections import Counter
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Self

import numpy as np

from iqm.pulse.builder import CircuitOperation, build_quantum_ops
from iqm.pulse.quantum_ops import QuantumOp, QuantumOpTable


def reorder(A: np.ndarray, perm: list[int]) -> np.ndarray:
    """Permutes the order of the qubits in an n-qubit operator matrix.

    Args:
        A: Matrix of an operator acting on the state space of ``n`` qubits, to be reordered.
        perm: Permutation vector for the ``n`` qubits.
            ``perm[k]`` is the new index for the k:th qubit in the old ordering.
            For example, ``reorder(kron(A, B, C)), [2, 0, 1]) == kron(B, C, A)``.

    Returns:
        Reordered ``A``.

    """
    shape = A.shape
    assert shape[0] == shape[1]  # must be square
    n_qubits = round(np.log2(shape[0]))  # must be a power of two

    if set(perm) != set(range(n_qubits)):
        raise ValueError(f"Invalid {n_qubits}-qubit permutation: {perm}.")

    # invert the permutation, iperm[k] is the old index of the k:th qubit in the new ordering
    iperm = np.argsort(perm)
    total_perm = np.r_[iperm, iperm + n_qubits]
    dims = [2] * n_qubits
    total_dims = dims * 2

    # big-endian ordering
    # reshape A into another tensor which has one index per qubit, permute dimensions,
    # back into a tensor with the original number of indices
    return A.reshape(total_dims).transpose(total_perm).reshape(shape)


@lru_cache(maxsize=1000)
def reshape_unitary(unitary: tuple[tuple[float, ...], ...], indices: tuple[int, ...], n_qubits: int) -> np.ndarray:
    """Extend a unitary propagator to act on a larger system.

    Given a unitary matrix acting on N qubits, indices of N qubits in a larger Hilbert space, and the number
    of qubits in that Hilbert space, calculate a unitary acting on the larger Hilbert space, which acts as
    the given unitary on the N selected qubits and as identity on the others.

    Args:
        unitary: original unitary as a nested tuple for hashing & caching purposes.
        indices: qubit indices in the larger Hilbert space, all in ``range(num_qubits)``
        n_qubits: number of qubits in the larger Hilbert space

    Returns:
        Unitary matrix acting as ``unitary`` on the selected qubits in the larger Hilbert space.

    """
    unitary_array = np.array(unitary)
    n_indices = len(indices)
    dim = 1 << n_indices
    if unitary_array.shape != (dim, dim):
        raise ValueError(f"Unitary does not have the correct dimension {dim}.")

    permutation = list(indices) + [i for i in range(n_qubits) if i not in indices]
    big_unitary = np.kron(unitary_array, np.eye(1 << (n_qubits - n_indices)))
    return reorder(big_unitary, permutation)


def get_unitary_from_op(op: CircuitOperation, table: QuantumOpTable, qubits: list[str]) -> np.ndarray:
    """Unitary matrix representing an operation within the context of the circuit.

    First, fetches the unitary corresponding to the correct operation from the QuantumOpTable. If that unitary
    is a function, gets the matrix by calling the function with values of quantum operation params stored
    in the operation. Checks whether the size of the unitary is correct. Then, optionally extends that unitary to act
    on the Hilbert space of the entire circuit.

    Args:
        op: quantum operation instance
        table: registered quantum operations
        qubits: qubits of the whole circuit, in big-endian order

    Returns:
        Unitary matrix representing ``op`` on the Hilbert space of the circuit.

    """
    try:
        parent_op = table[op.name]
    except KeyError as excp:
        raise KeyError(f"The operation {op.name} cannot be found in the QuantumOpTable.") from excp

    if parent_op.arity == 0:
        return np.eye(2 ** len(qubits))

    base_unitary_gen = parent_op.unitary
    if base_unitary_gen is None:
        raise ValueError(f"The operation {op.name} is not unitary or not fully defined over the computational space.")

    base_unitary = base_unitary_gen(**op.args)

    # args to reshape_unitary need to be hashable to enable caching
    locus_indices = tuple(qubits.index(qb) for qb in op.locus)
    tuple_unitary = tuple(map(tuple, base_unitary))

    return reshape_unitary(tuple_unitary, locus_indices, len(qubits))


def get_unitary_from_circuit(
    circuit: list[CircuitOperation], table: QuantumOpTable | None = None, qubit_names: list[str] | None = None
) -> np.ndarray:
    """Calculate the overall unitary implemented by a sequence of CircuitOperations.

    Iterate through the list of operations, skipping over barrier operations, and calculate the unitary
    for each operation, and then calculate the matrix product of all of them. The unitary definition must be present
    in the QuantumOpTable given as the second argument.

    Args:
        circuit: list of CircuitOperations in order
        table: Table of all registered quantum ops.
        qubit_names: Optionally, the ordering of the qubits.

    Returns:
        Array describing the action of the circuit in big endian convention.

    """
    table = table or build_quantum_ops({})
    qubit_names = qubit_names or _get_qubit_order_from_circuit(circuit)
    unitary = np.eye(2 ** len(qubit_names))  # in case of empty list
    for operation in circuit:
        unitary = get_unitary_from_op(operation, table, qubit_names) @ unitary

    return unitary


def _get_qubit_order_from_circuit(circuit: list[CircuitOperation]):  # noqa: ANN202
    """Get all qubits which are in the circuit in order."""
    # unique items, in the same order
    return list(dict.fromkeys(qb for op in circuit for qb in op.locus))


class CircuitOperationList(list):
    """List of :class:`.CircuitOperation` objects representing a quantum circuit.

    The class is used to work with CircuitOperations directly. It is mostly meant as
    convenience to enable easy creation of circuits, calculations of their properties, and mapping them onto physical
    qubits. In addition to the circuit contents, this class has two important attributes:
    :attr:`qubits` and :attr:`table`.

    :attr:`qubits` defines the list of qubits which are allowed to be in the loci of all the
    CircuitOperations present in the list. Think about it as Qiskit's QuantumRegister.

    :attr:`table` is a :class:`.QuantumOpTable`, which contains all the
    :class:`.QuantumOp` s which are allowed in the circuit. In most cases, the table is
    simply taken to contain all the default operations defined in :mod:`iqm.pulse`.
    When you use this class with a :class:`.ScheduleBuilder`, it is good practice to set
    ``table = builder.op_table``. The QuantumOpTable is mutable, so any additional registered
    gates can automatically be usable in any CircuitOperationList associated with that
    ScheduleBuilder instance.

    The fundamental use of the class would be to first define a new instance:

    .. code-block:: python

        circuit = CircuitOperationList(num_qubits=2)

    The ``num_qubits`` parameter populates the :attr:`qubits` attribute with qubits QB1-QBn,
    in this case ``['QB1', 'QB2']``.

    Alternatively, you can provide ``qubits`` directly:

    .. code-block:: python

        circuit = CircuitOperationList(qubits=['QB1', 'QB2'])

    To add your own QuantumOpTable, initialize like this:

    .. code-block:: python

        circuit = CircuitOperationList(num_qubits=2, table=my_table)

    Remembering that the table is mutable.

    If you already have a list of CircuitOperations, you can initialize with it:

    .. code-block:: python

        circuit = CircuitOperationList(circuit_ops, table=my_table)
        circuit.find_qubits()

    Calling the :meth:`find_qubits` method populates the :attr:`qubits` attribute with the qubits found in loci of
    the operations in the original circuit. If the list is empty, it will set :attr:`qubits` to an empty list,
    which most of the time is not what you want to do.

    The class has the ``__add__``, ``__mul__`` and ``__getitem__`` methods redefined, which means
    ``circuit * 3``, ``circuit1 + circuit2`` and ``circuit[0:4]`` will produce a CircuitOperationList
    with the same :attr:`qubits` and :attr:`table` attributes as the original.

    To add a ``prx`` operation to the list, call:

    .. code-block:: python

        circuit.add_op('prx', [0], angle, phase, impl_name='drag_crf')

    The class also has shortcut methods defined, so the above can be shortened to

    .. code-block:: python

        circuit.prx(angle, phase, 0, impl_name='drag_crf')

    which is exactly the same syntax as in Qiskit, with the addition of the implementation name
    which usually does not need to be used. The names of the shortcut methods are taken from the
    attached ``table`` at init. All the operations with non-zero arity
    will be added as shortcuts.

    If all the operations in the circuit are unitary, you can calculate the unitary propagator of
    the entire circuit by calling:

    .. code-block:: python

        U = circuit.get_unitary()

    The dimension of the unitary will always be defined by the :attr:`qubits` attribute. In particular, if your circuit
    contains 3 qubits, ``'QB1', 'QB2', 'QB3'``, but you only add gates to the first two, the resulting unitary will
    still be an 8x8 matrix, corresponding to the three qubits ``'QB1', 'QB2', 'QB3'``, in the big endian convention.
    With no operations affecting ``'QB3'``, the action of the unitary on this qubit is identity.

    To map the circuit onto physical qubits, all you need to do is call:

    .. code-block:: python

        physical_circuit = circuit.map_loci(physical_qubits)

    This will create a copy of the circuit, with all the placeholder qubits replaced by the physical qubits, with the
    order defined by the :attr:`qubits` attribute. For example, if ``qubits = ['QB1', 'Alice', 'ZZZ']``, and
    ``physical_qubits = ['QB2', 'QB5', 'QB10']``, all occurrences of ``'QB1'`` will be mapped to ``'QB2'``, ``'Alice'``
    to ``'QB5'`` and ``'ZZZ'`` to ``'QB10'``. The original circuit is not modified, so you can create many copies with
    different physical qubits, which is helpful when running parallel experiments on a large chip.

    Args:
        contents: Circuit operations to initialize the circuit with. Can be left out.
        qubits: Qubits allowed to be used in operation loci in the circuit.
        num_qubits: Number of qubits in the circuit, will initialize ``qubits`` with ``['QB1', 'QB2', ...]``.
            Ignored if ``qubits`` is given.
        table: Allowed quantum operations.

    """

    qubits: list[str] = []

    def __init__(
        self,
        contents: Iterable[CircuitOperation] = (),
        *,
        qubits: list[str] | None = None,
        num_qubits: int = 0,
        table: QuantumOpTable | None = None,
    ):
        list.__init__(self, contents)
        if qubits:
            self.qubits = qubits
        else:
            self.add_qubits(num_qubits)
        self.table: QuantumOpTable = table or build_quantum_ops({})
        for op_name in self.table:
            if self.table[op_name].arity:
                self._set_specific_operation_shortcut(op_name)

    def __getitem__(self, item) -> CircuitOperationList | CircuitOperation:  # type: ignore[override]  # type: ignore[override]  # type: ignore[override]  # noqa: ANN001
        """For the builtin list, this method is used both for accessing a single element: ``mylist[0]`` and accessing
        a slice: ``mylist[1:3]``. The latter should generate a new CircuitOperationList, so we override the method to
        ensure that it does.
        """
        result = list.__getitem__(self, item)
        if isinstance(result, list):
            new = CircuitOperationList(result, qubits=self.qubits, table=self.table)
            return new

        return result

    def __add__(self, other) -> CircuitOperationList:  # noqa: ANN001
        new = CircuitOperationList(list.__add__(self, other), qubits=self.qubits, table=self.table)

        return new

    def __mul__(self, other) -> CircuitOperationList:  # noqa: ANN001
        new = CircuitOperationList(list.__mul__(self, other), qubits=self.qubits, table=self.table)
        return new

    def find_qubits(self) -> None:
        """Set attribute qubits to qubits in the loci of operations in the list."""
        self.qubits = list(dict.fromkeys(qb for operation in self for qb in operation.locus))

    def add_qubits(self, n: int) -> None:
        """Adds generic placeholder qubits from 1 to n."""
        self.qubits = [f"QB{i + 1}" for i in range(n)]

    def get_unitary(self, qubit_names: list[str] | None = None) -> np.ndarray:
        """Calculate the overall unitary implemented by a sequence of CircuitOperations.

        Args:
            self: list of CircuitOperations in order
            qubit_names: Optionally, the ordering of the qubits.

        Returns:
            Array describing the action of the circuit in big endian convention.

        """
        qubit_names = qubit_names or self.qubits
        return get_unitary_from_circuit(self, self.table, qubit_names)

    def add_op(
        self,
        name: str,
        locus_indices: Sequence[int],
        *args,
        impl_name: str | None = None,
    ) -> None:
        """Adds a new :class:`CircuitOperation` to the circuit.

        Appends a new :class:`~iqm.pulse.builder.CircuitOperation` at the end of the list. The
        :class:`CircuitOperation` is created using a :class:`~iqm.pulse.quantum_ops.QuantumOp` name from the
        QuantumOpTable attached to the CircuitOperationList. The locus of that
        :class:`~iqm.pulse.builder.CircuitOperation` is built from the qubits stored in :attr:`qubits`, by selecting
        the qubits at indices given by ``locus_indices``. For example, if :attr:`qubits` is ``['QB1', 'QB2',
        'QB4']``, and the ``locus_indices`` is ``[2, 1]``, the locus of the new
        :class:`~iqm.pulse.builder.CircuitOperation` will be ``('QB4', 'QB2')``. All arguments for the values of the
        params of the requested :class:`~iqm.pulse.quantum_ops.QuantumOp` need to be provided.

        Args:
            name: Name of the :class:`QuantumOp` which will generate a new :class:`CircuitOperation`.
            locus_indices: Indices of the qubits in the attribute .qubits which will become the locus of the operation.
            args: Any arguments the CircuitOperation needs, must correspond to the params of the :class:`QuantumOp`.
            impl_name: Name of the implementation to use when converting the :class:`CircuitOperation` into
                a :class:`Timebox` later.

        """
        qubit_names = self.qubits
        if name not in self.table:
            raise KeyError(f"QuantumOp with name {name} is not in the gate definitions table.")

        try:
            locus = tuple(qubit_names[idx] for idx in locus_indices)
        except IndexError as e:
            raise IndexError(
                "To add new operations in this way, make sure the attribute 'qubits' has enough qubits. "
                "It can also be automatically populated by calling the find_qubits() method if there"
                "already are operations, or set_qubits(n) if the list is empty."
            ) from e

        params = self.table[name].params
        if len(params) != len(args):
            raise TypeError(
                f"Operation {name} has the following arguments: {tuple(params)}, but {len(args)} values were provided."
            )
        # Convert int args to floats, so that one can pass e.g. 0 instead of 0.0 for the arg.
        # Note that this conversion is not exact if the int value is too large, but such values are not realistic
        # for the current parameter types.
        args = tuple(
            float(arg) if (float in param_types and int not in param_types) and isinstance(arg, int) else arg
            for arg, param_types in zip(args, params.values())
        )
        new_op = CircuitOperation(name=name, args=dict(zip(params, args)), implementation=impl_name, locus=locus)
        new_op.validate(self.table)
        self.append(new_op)

    # define method for barrier separately as it is meta and behaves differently
    def barrier(self, *locus_indices) -> None:
        """Add barrier to the circuit"""
        indices = list(locus_indices) if locus_indices else list(np.arange(len(self.qubits), dtype=int))
        self.add_op("barrier", indices)

    def compose(
        self,
        other,  # noqa: ANN001
        locus_indices: list[int] | None = None,
    ) -> Self:
        """A safer way to add circuits together, but will probably take time.

        All the :class:`~iqm.pulse.builder.CircuitOperation` s from the ``'other'`` list are appended to the end of
        this list. The wire ``k`` of the second circuit is connected to wire ``locus_indices[k]`` of the first. This
        is achieved by mapping the locus of each operation in the second circuit onto the qubits of the first.

        For example, if the :attr:`qubits` of the first list are ``['QB1', 'QB2']``, the second list has
        ``['QB3', 'QB4']``, and the locus_indices argument is ``[1,0]``, all the operations in the second list will have
        their ``'QB3'`` mapped to ``'QB2'`` and ``'QB4'`` mapped to ``'QB1'``.

        Args:
            other: Second CircuitOperationList. Must have less or equal qubits than this one.
            locus_indices: Indices of the qubits in this CircuitOperationList onto which the qubits in the second
                circuit ar mapped.

        Returns:
            Self, with new operations added.

        """
        qubit_names1 = self.qubits
        qubit_names2 = other.qubits

        locus_indices = locus_indices or list(range(len(self.qubits)))

        if len(locus_indices) < len(qubit_names2):
            raise IndexError(
                f"There are {len(qubit_names2)} qubits in the second circuit, but provided mapping is"
                f" {locus_indices}. Make sure the length of the mapping is equal or longer to number of qubits."
            )
        locus_indices = locus_indices[: len(qubit_names2)]

        new_locus = [qubit_names1[idx] for idx in locus_indices]

        new_circuit = other.map_loci(new_locus, make_circuit=False)

        self.extend(new_circuit)

        return self

    def count_ops(self) -> Counter:
        """Count each type of operation in the circuit.

        Returns:
            Counter mapping operation names to numbers of times they occur in the circuit.

        """
        return Counter(operation.name for operation in self)

    def map_loci(
        self,
        locus: list[str] | None,
        make_circuit: bool = True,
    ) -> CircuitOperationList | list[CircuitOperation]:
        """Creates a new list of :class:`CircuitOperation` s with locus mapped onto physical qubits.

        Creates a fresh list of fresh :class:`~iqm.pulse.builder.CircuitOperation` s with fresh arguments. If
        ``locus`` is provided, it needs to have the same length as the total number of qubits across the circuit,
        and the qubits will then be mapped onto the new locus. If it is not provided, this is identical to a deepcopy
        of the original list.

        Args:
            locus: List of new qubits to replace the qubits in the loci of the operations in the circuit.
            make_circuit: If True, creates a :class:`CircuitOperationList`. If False, it is just a list.

        Returns:
            New CircuitOperationList with loci mapped onto new locus.

        """
        logical_locus = self.qubits
        locus = locus or logical_locus
        if len(locus) != len(set(locus)):
            raise ValueError("Repeated locus elements.")
        if len(logical_locus) != len(locus):
            raise IndexError(
                "The number of qubits in the new locus must be equal to the number of qubits in the old locus."
            )
        mapping = dict(zip(logical_locus, locus))
        new_circuit = CircuitOperationList(qubits=locus, table=self.table) if make_circuit else []
        for inst in self:
            new_args = inst.args.copy()
            # The classically conditioned operation has a key which specifies which qubit to receive feedback from.
            # This should be correctly mapped.
            # TODO: remove this exception after Pulla/CoCos compilation handles feedback sources in a more regular way
            if "feedback_qubit" in new_args:
                new_args["feedback_qubit"] = mapping[new_args["feedback_qubit"]]
            new_circuit.append(
                CircuitOperation(
                    name=inst.name,
                    locus=tuple(mapping[qb] for qb in inst.locus),
                    args=new_args,
                    implementation=inst.implementation,
                )
            )

        return new_circuit

    def _set_specific_operation_shortcut(self, name: str) -> None:
        """Add the convenience methods for adding new operations, based on the default :class:`QuantumOpTable`."""
        op = self.table[name]
        num_params = len(op.params)
        arity = op.arity

        def _add_specific_op(  # noqa: ANN202
            self,  # noqa: ANN001
            *args_and_locus,
            impl_name: str | None = None,
        ):
            if len(args_and_locus) != num_params + arity:
                raise TypeError(
                    f"The operation {name} requires {arity} locus/qubit indices and {num_params} additional"
                    f" arguments. A total of {len(args_and_locus)} were provided."
                )
            args = args_and_locus[:num_params]
            locus_indices = [int(idx) for idx in args_and_locus[num_params:]]
            self.add_op(name, locus_indices, *args, impl_name=impl_name)

        _add_specific_op.__doc__ = (
            f"Add a new operation {name}. The first {num_params} arguments are arguments to"
            f" the operation, if any, eg. angle and phase. The last {arity} arguments are indices"
            f" of the qubits the operation acts on. The keyword argument impl_name can be provided"
            f" optionally to fix the implementation of the operation in iqm-pulse."
        )
        setattr(CircuitOperationList, name, _add_specific_op)


@dataclass
class Circuit:
    """Quantum circuit.

    Used e.g. for client-server communication.

    Consists of a sequence of native quantum operations, each represented by an instance of the
    :class:`CircuitOperation` class.
    """

    name: str
    """name of the circuit"""
    instructions: tuple[CircuitOperation, ...]
    """operations comprising the circuit"""
    metadata: dict[str, Any] | None = None
    """optional metadata for the circuit"""

    def all_locus_components(self) -> set[str]:
        """Return the names of all locus components (typically qubits) in the circuit."""
        components: set[str] = set()
        for instruction in self.instructions:
            components.update(instruction.locus)
        return components

    def validate(self, supported_operations: dict[str, QuantumOp]) -> None:
        """Validate the circuit against the supported quantum operations.

        Args:
            supported_operations: mapping of supported quantum operation names to their definitions

        Raises:
            ValueError: circuit is not valid

        """
        self._validate_name()
        self._validate_instructions(supported_operations)

    def _validate_name(self) -> None:
        if not self.name:
            raise ValueError("A circuit should have a non-empty string for a name.")

    def _validate_instructions(self, supported_operations: dict[str, QuantumOp]) -> None:
        if not self.instructions:
            raise ValueError("Each circuit should have at least one instruction.")

        for instruction in self.instructions:
            if isinstance(instruction, CircuitOperation):
                instruction.validate(supported_operations)
            else:
                raise ValueError("Every instruction in a circuit should be of type <CircuitOperation>")
